import pya
from pya import *
from SiEPIC.utils import get_technology_by_name


class ebeam_taper(pya.PCellDeclarationHelper):
    """
    The PCell declaration for the strip waveguide taper.
    """

    def __init__(self):
        # Important: initialize the super class
        super(ebeam_taper, self).__init__()
        TECHNOLOGY = get_technology_by_name("EBeam")

        # Button to launch separate window
        self.param(
            "documentation", self.TypeCallback, "Open documentation in web browser"
        )
        #    self.param("simulation", self.TypeCallback, "Launch simulation GUI")

        # declare the parameters
        self.param("silayer", self.TypeLayer, "Si Layer", default=TECHNOLOGY["Si"])
        self.param("wg_width1", self.TypeDouble, "Waveguide Width1", default=0.5)
        self.param("wg_width2", self.TypeDouble, "Waveguide Width2", default=3)
        self.param("wg_length", self.TypeDouble, "Waveguide Length", default=10)
        self.param(
            "pinrec", self.TypeLayer, "PinRec Layer", default=TECHNOLOGY["PinRec"]
        )
        self.param(
            "devrec", self.TypeLayer, "DevRec Layer", default=TECHNOLOGY["DevRec"]
        )
        # hidden parameters, can be used to query this component:
        self.param(
            "p1",
            self.TypeShape,
            "DPoint location of pin1",
            default=Point(-10000, 0),
            hidden=True,
            readonly=True,
        )
        self.param(
            "p2",
            self.TypeShape,
            "DPoint location of pin2",
            default=Point(0, 10000),
            hidden=True,
            readonly=True,
        )

    def callback(self, layout, name, states):
        """Callback for PCell, to launch documentation viewer
        https://www.klayout.de/doc/code/class_PCellDeclaration.html#method9
        """
        if name == "documentation":
            url = "https://github.com/SiEPIC/SiEPIC_EBeam_PDK/blob/master/Documentation/Taper/Summary_Taper.pdf"
            import webbrowser

            webbrowser.open_new(url)

    def display_text_impl(self):
        # Provide a descriptive text for the cell
        return (
            "ebeam_taper(R="
            + ("%.3f-%.3f-%.3f" % (self.wg_width1, self.wg_width2, self.wg_length))
            + ")"
        )

    def can_create_from_shape_impl(self):
        return False

    def produce(self, layout, layers, parameters, cell):
        """
        coerce parameters (make consistent)
        """
        self._layers = layers
        self.cell = cell
        self._param_values = parameters
        self.layout = layout
        shapes = self.cell.shapes

        # cell: layout cell to place the layout
        # LayerSiN: which layer to use
        # w: waveguide width
        # length units in dbu

        # fetch the parameters
        dbu = self.layout.dbu
        ly = self.layout

        LayerSi = self.silayer
        LayerSiN = self.silayer_layer
        LayerPinRecN = ly.layer(self.pinrec)
        LayerDevRecN = ly.layer(self.devrec)

        w1 = int(round(self.wg_width1 / dbu))
        w2 = int(round(self.wg_width2 / dbu))
        length = int(round(self.wg_length / dbu))

        pts = [
            Point(0, -w1 / 2),
            Point(0, w1 / 2),
            Point(length, w2 / 2),
            Point(length, -w2 / 2),
        ]
        shapes(LayerSiN).insert(Polygon(pts))

        # Create the pins on the waveguides, as short paths:
        from SiEPIC._globals import PIN_LENGTH as pin_length

        # Pin on the left side:
        p1 = [Point(pin_length / 2, 0), Point(-pin_length / 2, 0)]
        p1c = Point(0, 0)
        self.set_p1 = p1c
        self.p1 = p1c
        pin = Path(p1, w1)
        shapes(LayerPinRecN).insert(pin)
        t = Trans(Trans.R0, 0, 0)
        text = Text("opt1", t)
        shape = shapes(LayerPinRecN).insert(text)
        shape.text_size = 0.4 / dbu

        # Pin on the right side:
        p2 = [Point(length - pin_length / 2, 0), Point(length + pin_length / 2, 0)]
        p2c = Point(length, 0)
        self.set_p2 = p2c
        self.p2 = p2c
        pin = Path(p2, w2)
        shapes(LayerPinRecN).insert(pin)
        t = Trans(Trans.R0, length, 0)
        text = Text("opt2", t)
        shape = shapes(LayerPinRecN).insert(text)
        shape.text_size = 0.4 / dbu
        shape.text_halign = 2

        # Create the device recognition layer -- make it 1 * wg_width away from the waveguides.
        path = Path([Point(0, 0), Point(length, 0)], w2 + w1 * 2)
        shapes(LayerDevRecN).insert(path.simple_polygon())

        # Compact model information
        t = Trans(Trans.R0, w1 / 10, 0)
        text = Text("Lumerical_INTERCONNECT_library=None", t)
        shape = shapes(LayerDevRecN).insert(text)
        shape.text_size = length / 100
        t = Trans(Trans.R0, length / 10, w1 / 4)
        text = Text("Component=ebeam_taper", t)
        shape = shapes(LayerDevRecN).insert(text)
        shape.text_size = length / 100
        t = Trans(Trans.R0, length / 10, w1 / 2)
        text = Text(
            "Spice_param:wg_width1=%.3fu wg_width2=%.3fu wg_length=%.3fu"
            % (self.wg_width1, self.wg_width2, self.wg_length),
            t,
        )
        shape = shapes(LayerDevRecN).insert(text)
        shape.text_size = length / 100

        return (
            "ebeam_taper("
            + ("%.3f-%.3f-%.3f" % (self.wg_width1, self.wg_width2, self.wg_length))
            + ")"
        )



import pya
from pya import *
from SiEPIC.utils import get_technology_by_name


class Waveguide_SWG_to_Strip(pya.PCellDeclarationHelper):
    """
    Input: length, period_strip, period_swg, wg_width_strip, wg_width_swg, duty_strip, duty_swg
    continuously (linearly) variable period, width, duty along the length

    Updated by Lukas Chrostowski, Leanne Dias, Connor Mosquera 2020/05
    """

    def __init__(self):
        # Important: initialize the super class
        super(Waveguide_SWG_to_Strip, self).__init__()
        TECHNOLOGY = get_technology_by_name("EBeam")

        # declare the parameters
        self.param("wavelength", self.TypeDouble, "Wavelength", default=1310)
        self.param("fishbone", self.TypeBoolean, "Fishbone", default=False)
        self.param("length", self.TypeDouble, "Waveguide length", default=10.0)
        self.param(
            "taper_fraction",
            self.TypeDouble,
            "Strip taper length fraction (0 to 1)",
            default=1,
        )
        self.param(
            "period_strip",
            self.TypeDouble,
            "SWG Period at strip end (microns)",
            default=0.200,
        )
        self.param(
            "period_swg",
            self.TypeDouble,
            "SWG Period at SWG end (microns)",
            default=0.200,
        )
        self.param(
            "wg_width_strip",
            self.TypeDouble,
            "Strip Waveguide width at the taper end (microns)",
            default=0.5,
        )
        #    self.param("wg_width_swg_taperend", self.TypeDouble, "Waveguide width at SWG taper end (microns)", default = 0.4)
        self.param(
            "wg_width_swg",
            self.TypeDouble,
            "SWG Waveguide at end (microns)",
            default=0.5,
        )
        self.param(
            "wg_width_taper",
            self.TypeDouble,
            "Nanotaper width at taper end (microns)",
            default=0.06,
        )
        self.param(
            "duty_strip",
            self.TypeDouble,
            "SWG duty cycle at strip end (0 to 1)",
            default=0.500,
        )
        self.param(
            "duty_swg",
            self.TypeDouble,
            "SWG duty cycle at SWG end (0 to 1)",
            default=0.700,
        )
        self.param("layer", self.TypeLayer, "Layer", default=TECHNOLOGY["Si"])
        self.param(
            "pinrec", self.TypeLayer, "PinRec Layer", default=TECHNOLOGY["PinRec"]
        )
        self.param(
            "devrec", self.TypeLayer, "DevRec Layer", default=TECHNOLOGY["DevRec"]
        )

    def display_text_impl(self):
        # Provide a descriptive text for the cell
        return "Waveguide_SWG_to_Strip_%s-%.3f-%.3f-%.3f" % (
            self.length,
            self.period_swg,
            self.wg_width_swg,
            self.duty_swg,
        )

    def coerce_parameters_impl(self):
        pass

    def can_create_from_shape(self, layout, shape, layer):
        return False

    def produce_impl(self):
        debug = False
        # Layout and layers:
        dbu = self.layout.dbu
        ly = self.layout
        shapes = self.cell.shapes

        LayerSi = self.layer
        LayerSiN = ly.layer(LayerSi)
        LayerSiSPN = ly.layer(LayerSi)
        LayerPinRecN = ly.layer(self.pinrec)
        LayerDevRecN = ly.layer(self.devrec)

        # fetch the PCell parameters, and calculate a few things
        from SiEPIC.extend import to_itype

        wg_width_swg = to_itype(self.wg_width_swg, dbu)
        wg_width_strip = to_itype(self.wg_width_strip, dbu)
        length = to_itype(self.length, dbu)
        w1 = to_itype(self.wg_width_taper, dbu)
        w2 = to_itype(self.wg_width_strip, dbu)

        # Determine the period such that the waveguide length is as desired.  Slight adjustment to period
        N_boxes = int(
            round(self.length / (self.period_swg + self.period_strip) * 2.0) - 0.5
        )
        grating_period = self.length / (N_boxes) / dbu
        if debug:
            print("N boxes: %s, grating_period: %s" % (N_boxes, grating_period))

        # taper length, minus last SWG block
        # taper_length = int(round(self.taper_fraction * self.length / dbu))
        taper_length = (N_boxes - 1) * grating_period

        # Find out what is the delta width required between
        # a) the SWG end and b) the SWG with nanotaper
        # Done by simulating the neff versus SWG waveguide width, with an without a 60 nm
        w = wg_width_swg
        if self.fishbone:
            dw = 0
        elif self.wavelength == 1310:
            dw = -3.3736e-5 * w * w + 0.0618752 * w + 3.88577
        else:
            dw = 0
        wg_width_swg_taperend = wg_width_swg - dw

        # Draw the SWG waveguide:

        x = -self.period_swg * self.duty_swg / 2 / dbu
        for i in range(0, N_boxes):
            local_duty = (
                1.0 * (N_boxes - i) / N_boxes * self.duty_swg
                + 1.0 * i / N_boxes * self.duty_strip
            )
            local_period = to_itype(
                (
                    1.0 * (N_boxes - i) / N_boxes * self.period_swg
                    + 1.0 * i / N_boxes * self.period_strip
                ),
                dbu,
            )
            local_wg_width = (
                1.0 * (N_boxes - i) / N_boxes * wg_width_swg_taperend
                + 1.0 * i / N_boxes * wg_width_strip
            )
            if (i == 0) | (i == N_boxes):
                if debug:
                    print(
                        "local_duty: %s, local_period: %s, local_wg_width: %s"
                        % (local_duty, local_period, local_wg_width)
                    )
            local_box_width = int(round(local_period * local_duty))
            #      x = int(round((i * local_period - local_box_width/2)))
            if i != 0:
                box1 = Box(
                    x, -local_wg_width / 2, x + local_box_width, local_wg_width / 2
                )
            else:
                # Last SWG, that is a bit larger than the one connected to the taper
                box1 = Box(x, -wg_width_swg / 2, x + local_box_width, wg_width_swg / 2)
            shapes(LayerSiN).insert(box1)
            x = x + int(round((local_period)))

        # Taper
        if self.fishbone:
            taper_length = length
        pts = [
            Point(length - taper_length, -w1 / 2),
            Point(length - taper_length, w1 / 2),
            Point(length, w2 / 2),
            Point(length, -w2 / 2),
        ]
        shapes(LayerSiN).insert(Polygon(pts))

        # Pins on the waveguide:

        from SiEPIC._globals import PIN_LENGTH as pin_length

        w = to_itype(self.wg_width_swg, dbu)
        t = Trans(Trans.R0, 0, 0)
        pin = Path([Point(pin_length / 2, 0), Point(-pin_length / 2, 0)], w)
        pin_t = pin.transformed(t)
        shapes(LayerPinRecN).insert(pin_t)
        text = Text("pin1", t)
        shape = shapes(LayerPinRecN).insert(text)
        shape.text_size = 0.4 / dbu

        w = to_itype(max(self.wg_width_strip, self.wg_width_taper), dbu)
        t = Trans(Trans.R0, length, 0)
        pin = Path([Point(-pin_length / 2, 0), Point(pin_length / 2, 0)], w)
        pin_t = pin.transformed(t)
        shapes(LayerPinRecN).insert(pin_t)
        text = Text("pin2", t)
        shape = shapes(LayerPinRecN).insert(text)
        shape.text_size = 0.4 / dbu
        shape.text_halign = 2

        """
    # Compact model information
    t = Trans(Trans.R0, 0, 0)
    text = Text ('Lumerical_INTERCONNECT_library=Design kits/ebeam_v1.2', t)
    shape = shapes(LayerDevRecN).insert(text)
    shape.text_size = 0.1/dbu
    t = Trans(Trans.R0, length/10, 0)
    text = Text ('Component=NO_MODEL_AVAILABLE', t)
    shape = shapes(LayerDevRecN).insert(text)
    shape.text_size = 0.1/dbu
    t = Trans(Trans.R0, length/9, -w*2)
    text = Text \
      ('Spice_param:length=%.3fu period_swg=%.3fu period_strip=%.3fu wg_width_swg=%.3fu wg_width_strip=%.3fu duty_swg=%.3f duty_strip=%.3f ' %\
      (self.length, self.period_swg, (self.period_strip), self.wg_width_swg, self.wg_width_strip, self.duty_swg, self.duty_strip), t )
    shape = shapes(LayerDevRecN).insert(text)
    shape.text_size = 0.1/dbu
    blabla
    """

        # Create the device recognition layer -- make it 1 * wg_width away from the waveguides.
        points = [pya.Point(0, 0), pya.Point(length, 0)]
        path = pya.Path(points, w * 5)
        shapes(LayerDevRecN).insert(path.simple_polygon())



import pya
from pya import *
from SiEPIC.utils import get_technology_by_name


class Bezier_Bend(pya.PCellDeclarationHelper):
    """
    Input: bezier number, effective radius, waveguide width, layers
    draws a 90 degree bezier bend with a given bezier factor
    Usage: instantiate, and use transformations (rotation)

    This implementation is Python translation of the 90 degree Bezier bend ample script for Mentor Graphics Pyxis
      by Jonas Flueckiger and Lukas Chrostowski's
      influenced by www.tinaja.com/glib/cubemath.pdf implementation

    Author: Mustafa Hammood     mustafa@ece.ubc.ca
    """

    def __init__(self):
        # Important: initialize the super class
        super(Bezier_Bend, self).__init__()
        TECHNOLOGY = get_technology_by_name("EBeam")

        # declare the parameters
        self.param("bezier_num", self.TypeDouble, "Bezier factor", default=25)
        self.param(
            "eff_r", self.TypeDouble, "Effective bend radius (microns)", default=5
        )
        self.param(
            "wg_width", self.TypeDouble, "Waveguide width (microns)", default=0.500
        )
        self.param("layer", self.TypeLayer, "Layer", default=TECHNOLOGY["Si"])
        self.param(
            "pinrec", self.TypeLayer, "PinRec Layer", default=TECHNOLOGY["PinRec"]
        )
        self.param(
            "devrec", self.TypeLayer, "DevRec Layer", default=TECHNOLOGY["DevRec"]
        )

    #    self.param("textl", self.TypeLayer, "Text Layer", default = LayerInfo(10, 0))

    def display_text_impl(self):
        # Provide a descriptive text for the cell
        return "Bezier_Number_%.3f-%.3f" % (
            self.bezier_num / 1000,
            self.wg_width / 1000,
        )

    def coerce_parameters_impl(self):
        pass

    def can_create_from_shape(self, layout, shape, layer):
        return False

    def produce_impl(self):
        # fetch the parameters
        dbu = self.layout.dbu
        ly = self.layout

        LayerSi = self.layer
        LayerSiN = ly.layer(LayerSi)
        LayerPinRecN = ly.layer(self.pinrec)
        LayerDevRecN = ly.layer(self.devrec)

        from SiEPIC.utils import arc_bezier
        from SiEPIC._globals import PIN_LENGTH as pin_length

        wg_width = self.wg_width
        w = self.wg_width / dbu

        r = self.eff_r / dbu
        bezier = self.bezier_num / 100

        x = -r
        y = r

        points = arc_bezier(r, 0, 360, bezier)
        a = pya.Path(points, w).simple_polygon()
        d = [each for each in a.each_point()]
        pt_idx = int(len(d) / 2)
        ptA = d[pt_idx]
        ptB = d[pt_idx + 1]
        # print(pt_idx)
        # print(d)
        if abs(ptA.x - ptB.x) * dbu == wg_width:
            new_y = (ptA.y + ptB.y) / 2
            d[pt_idx].y = new_y
            d[pt_idx + 1].y = new_y
        elif abs(ptA.y - ptB.y) * dbu == wg_width:
            new_x = (ptA.x + ptB.x) / 2
            d[pt_idx].x = new_x
            d[pt_idx + 1].x = new_x
        # print(d)

        self.cell.shapes(LayerSiN).insert(Polygon(d))

        # Create the pins, as short paths:
        w = int(round(self.wg_width / dbu))
        r = int(round(self.eff_r / dbu))
        x = r
        y = r

        # Pin on the top side:
        p2 = [Point(0, y - pin_length / 2), Point(0, y + pin_length / 2)]
        p2c = Point(x, y)
        self.set_p2 = p2c
        self.p2 = p2c
        pin = Path(p2, w)
        self.cell.shapes(LayerPinRecN).insert(pin)
        t = Trans(Trans.R0, 0, y)
        text = Text("pin2", t)
        shape = self.cell.shapes(LayerPinRecN).insert(text)
        shape.text_size = 0.4 / dbu

        x = 0
        y = 0
        # Pin on the left side:
        p1 = [Point(pin_length / 2 - r, 0), Point(-pin_length / 2 - r, 0)]
        p1c = Point(x, 0)
        self.set_p1 = p1c
        self.p1 = p1c
        pin = Path(p1, w)
        self.cell.shapes(LayerPinRecN).insert(pin)
        t = Trans(Trans.R0, -r, 0)
        text = Text("pin1", t)
        shape = self.cell.shapes(LayerPinRecN).insert(text)
        shape.text_size = 0.4 / dbu

        # Create the device recognition layer -- make it 1 * wg_width away from the waveguides.
        points = arc_bezier(r, 270, 360, bezier)
        self.cell.shapes(LayerDevRecN).insert(pya.Path(points, w * 3))



import pya
from pya import *
from SiEPIC.utils import get_technology_by_name


class Bragg_Straight_from_file(pya.PCellDeclarationHelper):
    """
    by Rui Cheng 2020, UBC; modified by Lukas Chrostowski
    Description: Load Bragg grating waveguide vertices from a TXT file and draw

    Input: file with the format: (x1,y1,x2,y2;...)
    which can be generated in Matlab by:
      co = co / dbu;
      fid=fopen(file,'w');
      fprintf(fid,'(');
      fprintf(fid,'%d,%d;',round(co));
      fprintf(fid,')');
      fclose(fid);

    """

    def __init__(self):
        # Important: initialize the super class
        super(Bragg_Straight_from_file, self).__init__()
        TECHNOLOGY = get_technology_by_name("EBeam")

        # declare the parameters
        self.param("forder", self.TypeString, "Forder", default="")
        self.param("name", self.TypeString, "Name", default="Square_BP_B3.5_N2536_DW6")
        self.param("filetype", self.TypeString, "File type (txt)", default="txt")
        self.param("layer", self.TypeLayer, "Layer", default=TECHNOLOGY["Si"])
        self.param("port_w", self.TypeDouble, "Port Waveguide width", default=0.5)
        self.param(
            "pinrec", self.TypeLayer, "PinRec Layer", default=TECHNOLOGY["PinRec"]
        )
        self.param(
            "devrec", self.TypeLayer, "DevRec Layer", default=TECHNOLOGY["DevRec"]
        )
        self.param("textl", self.TypeLayer, "Text Layer", default=TECHNOLOGY["Text"])

    def display_text_impl(self):
        # Provide a descriptive text for the cell
        return "Bragg_Straight_from_file_%s" % (self.name)

    def coerce_parameters_impl(self):
        pass

    def can_create_from_shape(self, layout, shape, layer):
        return False

    def produce_impl(self):
        import os
        from SiEPIC.extend import to_itype

        # fetch the parameters
        dbu = self.layout.dbu
        ly = self.layout
        shapes = self.cell.shapes

        LayerSi = self.layer
        LayerSiN = ly.layer(LayerSi)
        LayerPinRecN = ly.layer(self.pinrec)
        LayerDevRecN = ly.layer(self.devrec)
        TextLayerN = ly.layer(self.textl)
        LayerPinRec = ly.layer(self.pinrec)

        file_path = os.path.join(self.forder, self.name + "." + self.filetype)
        # print('Bragg_Straight_from_file: input file: ' + file_path)

        if not (os.path.isfile(file_path)):
            error = "File not found: " + file_path
            text = Text(error, Trans(0, 0))
            shape = shapes(TextLayerN).insert(text)
            shape.text_size = 1 / dbu
            box = Box(0, 0, len(error) * shape.text_size, shape.text_size)
            shapes(LayerDevRecN).insert(box)
            return

        if "txt" in self.filetype:
            # Format: (x1,y1;x2,y2;...), where x and y values are integers
            text_file = open(file_path, "r")
            lines = text_file.readlines()
        else:
            error = "File type not available: " + self.filetype
            text = Text(error, Trans(0, 0))
            shape = shapes(TextLayerN).insert(text)
            shape.text_size = 1 / dbu
            box = Box(0, 0, len(error) * shape.text_size, shape.text_size)
            shapes(LayerDevRecN).insert(box)
            return

        p = SimplePolygon().from_s(lines[0])
        shapes(LayerSiN).insert(p)

        # Create the pins on the waveguides, as short paths:
        from SiEPIC._globals import PIN_LENGTH as pin_length

        bbox = p.bbox()
        y = (bbox.p1.y + bbox.p2.y) / 2
        leftx = bbox.left
        port_w = to_itype(self.port_w, dbu)
        pin = Path([Point(pin_length / 2, 0), Point(-pin_length / 2, 0)], port_w)

        t = Trans(Trans.R0, bbox.left, y)
        pin_t = pin.transformed(t)
        shapes(LayerPinRecN).insert(pin_t)
        text = Text("pin1", t)
        shape = shapes(LayerPinRecN).insert(text)
        shape.text_size = 0.4 / dbu

        t = Trans(Trans.R180, bbox.right, y)
        pin_t = pin.transformed(t)
        shapes(LayerPinRecN).insert(pin_t)
        text = Text("pin2", t)
        shape = shapes(LayerPinRecN).insert(text)
        shape.text_size = 0.4 / dbu

        # Create the device recognition layer -- make it 1 * wg_width away from the waveguides.
        box = Box(bbox.left, y - 3 / 2 * port_w, bbox.right, y + 3 / 2 * port_w)
        shapes(LayerDevRecN).insert(box)



from pya import *
import pya


class Wireguide(pya.PCellDeclarationHelper):
    """
    RIT Integrated Photonics Group of Stefan Preble (Karl McNulty)
    new wireguide class (essentially the metal path equivalent to a waveguide but no bend is allowed)
    """

    def __init__(self):
        # Important: initialize the super class
        super(Wireguide, self).__init__()
        # declare the parameters
        from SiEPIC.utils import get_technology_by_name

        TECHNOLOGY = get_technology_by_name("EBeam")
        self.param(
            "path",
            self.TypeShape,
            "Path",
            default=DPath([DPoint(0, 0), DPoint(10, 0), DPoint(10, 10)], 0.5),
        )
        self.radius = 0
        self.param("width", self.TypeDouble, "Width", default=0.5)
        self.adiab = False
        self.param("layers", self.TypeList, "Layers", default=["M2_router"])
        self.param("widths", self.TypeList, "Widths", default=[5])
        self.param("offsets", self.TypeList, "Offsets", default=[0])

    def display_text_impl(self):
        # Provide a descriptive text for the cell
        return "Wireguide_%s" % self.path

    def coerce_parameters_impl(self):
        return

    def can_create_from_shape_impl(self):
        return self.shape.is_path()

    def transformation_from_shape_impl(self):
        return Trans(Trans.R0, 0, 0)

    def parameters_from_shape_impl(self):
        self.path = self.shape.dpath

    def produce_impl(self):
        from SiEPIC.utils import (
            arc_xy,
            arc_bezier,
            angle_vector,
            angle_b_vectors,
            inner_angle_b_vectors,
        )
        import pya
        from SiEPIC.extend import to_itype

        # print("EBeam.Wireguide")

        from SiEPIC.utils import get_technology_by_name

        TECHNOLOGY = get_technology_by_name("EBeam")

        dbu = self.layout.dbu
        wg_width = to_itype(self.width, dbu)
        path = self.path.to_itype(dbu)

        if not (
            len(self.layers) == len(self.widths)
            and len(self.layers) == len(self.offsets)
            and len(self.offsets) == len(self.widths)
        ):
            raise Exception(
                "There must be an equal number of layers, widths and offsets"
            )
        path.unique_points()
        turn = 0
        for lr in range(0, len(self.layers)):
            layer = self.layout.layer(TECHNOLOGY[self.layers[lr]])

            width = to_itype(self.widths[lr], dbu)
            offset = to_itype(self.offsets[lr], dbu)

            pts = path.get_points()
            wg_pts = [pts[0]]
            for i in range(1, len(pts) - 1):
                turn = (
                    (angle_b_vectors(pts[i] - pts[i - 1], pts[i + 1] - pts[i]) + 90)
                    % 360
                    - 90
                ) / 90
                dis1 = pts[i].distance(pts[i - 1])
                dis2 = pts[i].distance(pts[i + 1])
                angle = angle_vector(pts[i] - pts[i - 1]) / 90
                pt_radius = self.radius
                # determine the radius, based on how much space is available
                if len(pts) == 3:
                    pt_radius = min(dis1, dis2, pt_radius)
                else:
                    if i == 1:
                        if dis1 <= pt_radius:
                            pt_radius = dis1
                    elif dis1 < 2 * pt_radius:
                        pt_radius = dis1 / 2
                    if i == len(pts) - 2:
                        if dis2 <= pt_radius:
                            pt_radius = dis2
                    elif dis2 < 2 * pt_radius:
                        pt_radius = dis2 / 2
                # wireguide bends:
                if self.adiab:
                    wg_pts += (
                        Path(
                            arc_bezier(
                                pt_radius,
                                270,
                                270
                                + inner_angle_b_vectors(
                                    pts[i - 1] - pts[i], pts[i + 1] - pts[i]
                                ),
                                self.bezier,
                                DevRec="DevRec" in self.layers[lr],
                                dbu=dbu,
                            ),
                            0,
                        )
                        .transformed(Trans(angle, turn < 0, pts[i]))
                        .get_points()
                    )
                else:
                    wg_pts += (
                        Path(
                            arc_xy(
                                -pt_radius,
                                pt_radius,
                                pt_radius,
                                270,
                                270
                                + inner_angle_b_vectors(
                                    pts[i - 1] - pts[i], pts[i + 1] - pts[i]
                                ),
                                DevRec="DevRec" in self.layers[lr],
                                dbu=dbu,
                            ),
                            0,
                        )
                        .transformed(Trans(angle, turn < 0, pts[i]))
                        .get_points()
                    )
            wg_pts += [pts[-1]]
            wg_pts = pya.Path(wg_pts, 0).unique_points().get_points()
            wg_polygon = Path(wg_pts, wg_width)
            self.cell.shapes(layer).insert(wg_polygon)  # insert the wireguide

            # if self.layout.layer(TECHNOLOGY['Wireguide']) == layer:
            waveguide_length = wg_polygon.area() / self.width * dbu**2

        pts = path.get_points()
        LayerPinRecN = self.layout.layer(TECHNOLOGY["PinRecM"])

        # insert pins to wireguide
        t1 = Trans(angle_vector(pts[0] - pts[1]) / 90, False, pts[0])
        self.cell.shapes(LayerPinRecN).insert(
            Path([Point(-50, 0), Point(50, 0)], wg_width).transformed(t1)
        )
        self.cell.shapes(LayerPinRecN).insert(Text("pin1", t1, 0.3 / dbu, -1))

        t = Trans(angle_vector(pts[-1] - pts[-2]) / 90, False, pts[-1])
        self.cell.shapes(LayerPinRecN).insert(
            Path([Point(-50, 0), Point(50, 0)], wg_width).transformed(t)
        )
        self.cell.shapes(LayerPinRecN).insert(Text("pin2", t, 0.3 / dbu, -1))

        LayerDevRecN = self.layout.layer(TECHNOLOGY["DevRec"])

        # Compact model information
        angle_vec = angle_vector(pts[0] - pts[1]) / 90
        halign = 0  # left
        angle = 0
        pt2 = pts[0]
        pt3 = pts[0]
        if angle_vec == 0:  # horizontal
            halign = 2  # right
            angle = 0
            pt2 = pts[0] + Point(0, wg_width)
            pt3 = pts[0] + Point(0, -wg_width)
        if angle_vec == 2:  # horizontal
            halign = 0  # left
            angle = 0
            pt2 = pts[0] + Point(0, wg_width)
            pt3 = pts[0] + Point(0, -wg_width)
        if angle_vec == 1:  # vertical
            halign = 2  # right
            angle = 1
            pt2 = pts[0] + Point(wg_width, 0)
            pt3 = pts[0] + Point(-wg_width, 0)
        if angle_vec == -1:  # vertical
            halign = 0  # left
            angle = 1
            pt2 = pts[0] + Point(wg_width, 0)
            pt3 = pts[0] + Point(-wg_width, 0)



print("SiEPIC-EBeam-PDK Python module: siepic_ebeam_pdk v0.4.22, KLayout technology: EBeam")

# Load the KLayout technology, when running in Script mode
import pya
import os

if not pya.Technology().has_technology("EBeam"):
    tech = pya.Technology().create_technology("EBeam")
    tech.load(os.path.join(os.path.dirname(os.path.realpath(__file__)), "EBeam.lyt"))

# then import all the technology modules
from . import pymacros
print(pymacros.__file__)
'''
from .pymacros import (
    SiEPIC_EBeam_Library,
    SiEPIC_EBeam_Library_SiN,
    SiEPIC_EBeam_Library_Dream,
    SiEPIC_EBeam_Library_Beta,
    SiEPIC_EBeam_Library_ANT,
)
'''

# display the registered libraries
print("Loaded technology libraries: %s" % pya.Library.library_names())



from pya import *
import pya


def linspace_without_numpy(low, high, length):
    step = (high - low) * 1.0 / length
    return [low + i * step for i in range(length)]


class Universal_GC(pya.PCellDeclarationHelper):
    """
    Universal Grating Coupler PCell implementation.
    Analytical design based on "Grating Coupler Design Based on Silicon-On-Insulator", Yun Wang (2013). Master's Thesis, University of British Columbia, Canada
    Some PCell implementation adapted from the SiEPIC_EBeam library by Dr. Lukas Chrostowski, University of British Columbia, Canada
    Orignal script written by Timothy Richards (Simon Fraser University, BC, Canada) and Adam DeAbreu (Simon Fraser University, BC, Canada)
    Modified to use UGC method presented in : Yun Wang, Jonas Flueckiger, Charlie Lin, Lukas Chrostowski, "Universal grating coupler design," Proc. SPIE 8915, Photonics North 2013, 89150Y (11 October 2013),
    Transfered from Mentor Graphics to KLayout by Connor Mosquera (University of British Columbia, Canada)

    """

    def __init__(self):
        # Important: initialize the super class
        super(Universal_GC, self).__init__()
        from SiEPIC.utils import get_technology_by_name

        TECHNOLOGY = get_technology_by_name("EBeam")
        self.TECHNOLOGY = TECHNOLOGY

        # declare the parameters
        self.param(
            "wavelength", self.TypeDouble, "Design Wavelength (micron)", default=1.55
        )
        self.param(
            "Si_thickness", self.TypeDouble, "SiN Thickness (micron)", default=0.4
        )
        self.param(
            "etch_depth",
            self.TypeDouble,
            "Etch Depth (micron)",
            default=0.4,
            hidden=True,
        )
        self.param("pol", self.TypeString, "Polarization", default="TE")
        self.param("n_t", self.TypeDouble, "Cladding Index", default=1.444)
        self.param("angle_e", self.TypeDouble, "Taper Angle (deg)", default=35.0)
        self.param(
            "grating_length", self.TypeDouble, "Grating Length (micron)", default=15.0
        )
        self.param(
            "taper_length", self.TypeDouble, "Taper Length (micron)", default=19.0
        )
        self.param("dc", self.TypeDouble, "Duty Cylce", default=0.5)
        self.param("t", self.TypeDouble, "Waveguide Width (micron)", default=0.75)
        self.param("theta_c", self.TypeDouble, "Insertion Angle (deg)", default=8.0)

        # Layer parameters
        self.param("layer", self.TypeLayer, "Layer", default=TECHNOLOGY["SiN"])
        self.param(
            "pinrec", self.TypeLayer, "PinRec Layer", default=TECHNOLOGY["PinRec"]
        )
        self.param(
            "devrec", self.TypeLayer, "DevRec Layer", default=TECHNOLOGY["DevRec"]
        )

    def display_text_impl(self):
        # Provide a descriptive text for the cell
        return "Universal_GC_%.1f-%.2f-%.2f-%.2f-%.2f-%.2f" % (
            self.wavelength,
            self.theta_c,
            self.dc,
            self.angle_e,
            self.taper_length,
            self.t,
        )

    #    return "temporary placeholder"

    def coerce_parameters_impl(self):
        pass

    def can_create_from_shape(self, layout, shape, layer):
        return False

    def produce_impl(self):
        # fetch the parameters
        dbu = self.layout.dbu
        ly = self.layout
        shapes = self.cell.shapes
        TECHNOLOGY = self.TECHNOLOGY

        LayerSi = self.layer
        LayerSiN = ly.layer(LayerSi)
        LayerSiSPN = ly.layer(LayerSi)
        LayerPinRecN = ly.layer(self.pinrec)
        LayerDevRecN = ly.layer(self.devrec)
        LayerTextN = TECHNOLOGY["Text"]

        ######## effective index function ##########

        def effective_index(
            wl=self.wavelength,
            etch_depth=self.etch_depth,
            Si_thickness=self.Si_thickness,
            n_t=self.n_t,
            pol=self.pol,
            dc=self.dc,
        ):
            from math import pi, tan, sqrt

            point = 1001
            n_0 = n_t
            n_1 = 0
            n_3 = 1.444
            if 0:  # Silicon
                n_2 = sqrt(
                    7.9874
                    + (3.68 * pow(3.9328, 2) * pow(10, 30))
                    / (
                        pow(3.9328, 2) * pow(10, 30)
                        - pow(2 * 3.14 * 3 * pow(10, 8) / (wl * pow(10, -6)), 2)
                    )
                )  # Silicon wavelength-dependant index of refraction
            if 1:  # Silicon Nitride
                n_2 = 2.0  # approximate
            delta = n_0 - n_3
            t = Si_thickness
            t_slot = t - etch_depth
            k_0 = 2 * pi / wl

            b_0 = linspace_without_numpy(0, 0, point - 1)
            te_0 = linspace_without_numpy(0, 0, point - 1)
            te_1 = linspace_without_numpy(0, 0, point - 1)
            tm_0 = linspace_without_numpy(0, 0, point - 1)
            tm_1 = linspace_without_numpy(0, 0, point - 1)
            h_0 = linspace_without_numpy(0, 0, point - 1)
            q_0 = linspace_without_numpy(0, 0, point - 1)
            p_0 = linspace_without_numpy(0, 0, point - 1)
            qbar_0 = linspace_without_numpy(0, 0, point - 1)
            pbar_0 = linspace_without_numpy(0, 0, point - 1)

            # calculating neff for the silicon layer
            if delta < 0:
                n_1 = n_3
            else:
                n_1 = n_0

            for ii in range(0, point - 1):
                b_0[ii] = (
                    n_1 * k_0 + (n_2 - n_1) * k_0 / (point - 1) * ii
                )  # copied from .ample UGC: should this be point-1?

                h_0[ii] = sqrt(abs(pow(n_2 * k_0, 2) - pow(b_0[ii], 2)))
                q_0[ii] = sqrt(abs(pow(b_0[ii], 2) - pow(n_0 * k_0, 2)))
                p_0[ii] = sqrt(abs(pow(b_0[ii], 2) - pow(n_3 * k_0, 2)))

                pbar_0[ii] = pow(n_2 / n_3, 2) * p_0[ii]
                qbar_0[ii] = pow(n_2 / n_0, 2) * q_0[ii]

            # calculating neff for TE mode
            if pol == "TE":
                for ii in range(0, point - 1):
                    te_0[ii] = tan(h_0[ii] * t) - (p_0[ii] + q_0[ii]) / h_0[ii] / (
                        1 - p_0[ii] * q_0[ii] / pow(h_0[ii], 2)
                    )
                    te_1[ii] = tan(h_0[ii] * t_slot) - (p_0[ii] + q_0[ii]) / h_0[ii] / (
                        1 - p_0[ii] * q_0[ii] / pow(h_0[ii], 2)
                    )

                abs_te_0 = [abs(x) for x in te_0]
                abs_te_1 = [abs(x) for x in te_1]
                index_TE_0 = abs_te_0.index(min(abs_te_0))
                index_TE_1 = abs_te_1.index(min(abs_te_1))
                nTE_0 = b_0[index_TE_0] / k_0
                nTE_1 = b_0[index_TE_1] / k_0
                nTE_1 = n_3

                while nTE_0 < 1.5 or nTE_0 > 3:
                    abs_te_0[index_TE_0] = 100
                    index_TE_0 = abs_te_0.index(min(abs_te_0))
                    nTE_0 = b_0[index_TE_0] / k_0

                """  
        while (nTE_1<1.5 or nTE_1>3):
          abs_te_1[index_TE_1] = 100
          index_TE_1 = abs_te_1.index(min(abs_te_1))
          nTE_1 = b_0[index_TE_1]/k_0
        """

                ne = dc * nTE_0 + (1 - dc) * nTE_1

            # calculating neff for TE mode
            elif pol == "TM":
                for ii in range(0, point - 1):
                    tm_0[ii] = tan(h_0[ii] * t) - (pbar_0[ii] + qbar_0[ii]) / h_0[
                        ii
                    ] / (1 - pbar_0[ii] * qbar_0[ii] / pow(h_0[ii], 2))
                    tm_1[ii] = tan(h_0[ii] * t_slot) - (pbar_0[ii] + qbar_0[ii]) / h_0[
                        ii
                    ] / (1 - pbar_0[ii] * qbar_0[ii] / pow(h_0[ii], 2))

                abs_tm_0 = [abs(x) for x in tm_0]
                abs_tm_1 = [abs(x) for x in tm_1]
                index_TM_0 = abs_tm_0.index(min(abs_tm_0))
                index_TM_1 = abs_tm_1.index(min(abs_tm_1))
                nTM_0 = b_0[index_TM_0] / k_0
                nTM_1 = b_0[index_TM_1] / k_0

                while nTM_0 < 1.5 or nTM_0 > 3:
                    abs_tm_0[index_TM_0] = 100
                    index_TM_0 = abs_tm_0.index(min(abs_tm_0))
                    nTM_0 = b_0[index_TM_0] / k_0

                while nTM_1 < 1.5 or nTM_1 > 3:
                    abs_tm_1[index_TM_1] = 100
                    index_TM_1 = abs_tm_1.index(min(abs_tm_1))
                    nTM_1 = b_0[index_TM_1] / k_0

                ne = dc * nTM_0 + (1 - dc) * nTM_1

            else:
                print("Please type TE or TM for polarization")

            return ne

        #####################################

        from math import pi, cos, sin, tan
        from SiEPIC.utils import points_per_circle

        lambda_0 = self.wavelength  ##um wavelength of light

        n_e = effective_index()
        ne_fiber = 1  # effective index of the mode in the air
        period = self.wavelength / (n_e - sin(pi / 180 * self.theta_c) * ne_fiber)

        # Geometry
        wh = period * self.dc  ##thick grating

        gc_number = int(round(self.grating_length / period))  ##number of periods
        e = self.n_t * sin((pi / 180) * self.theta_c) / n_e
        N = round(
            self.taper_length * (1 + e) * n_e / lambda_0
        )  ##allows room for the taper

        start = pi - (pi / 180) * self.angle_e / 2
        stop = pi + (pi / 180) * self.angle_e / 2

        # Draw coupler grating.
        for j in range(gc_number):
            # number of points in the arcs:
            # calculate such that the vertex & edge placement error is < 0.5 nm.
            #   see "SiEPIC_EBeam_functions - points_per_circle" for more details
            radius = N * lambda_0 / (n_e * (1 - e)) + j * period
            seg_points = int(
                points_per_circle(radius / dbu, dbu=dbu) / 360.0 * self.angle_e
            )  # number of points grating arc
            theta_up = []
            for m in range(seg_points + 1):
                theta_up = theta_up + [start + m * (stop - start) / seg_points]
            theta_down = theta_up[::-1]

            ##big one
            r_up = []
            r_down = []
            for k in range(len(theta_up)):
                r_up = r_up + [
                    N * lambda_0 / (n_e * (1 - e * cos(float(theta_up[k]))))
                    + j * period
                    + period * (1 - self.dc)
                ]
            r_down = r_up[::-1]

            xr = []
            yr = []
            for k in range(len(theta_up)):
                xr = xr + [r_up[k] * cos(theta_up[k])]
                yr = yr + [r_up[k] * sin(theta_up[k])]

            xl = []
            yl = []
            for k in range(len(theta_down)):
                xl = xl + [(r_down[k] + wh) * cos(theta_down[k])]
                yl = yl + [(r_down[k] + wh) * sin(theta_down[k])]

            x = xr + xl
            y = yr + yl

            pts = []
            for i in range(len(x)):
                pts.append(Point.from_dpoint(DPoint(x[i] / dbu, y[i] / dbu)))

            polygon = Polygon(pts)
            shapes(LayerSiN).insert(polygon)

        # Taper section
        r_up = []
        r_down = []
        for k in range(len(theta_up)):
            r_up = r_up + [N * lambda_0 / (n_e * (1 - e * cos(float(theta_up[k]))))]
        r_down = r_up[::-1]

        xl = []
        yl = []
        for k in range(len(theta_down)):
            xl = xl + [(r_down[k]) * cos(theta_down[k])]
            yl = yl + [(r_down[k]) * sin(theta_down[k])]

        yr = [self.t / 2.0, -self.t / 2.0]

        yl_abs = []
        for k in range(len(yl)):
            yl_abs = yl_abs + [abs(yl[k])]

        y_max = max(yl_abs)
        iy_max = yl_abs.index(y_max)

        L_o = (y_max - self.t / 2) / tan((pi / 180) * self.angle_e / 2)

        xr = [0, 0]

        x = xr + xl
        y = yr + yl

        pts = []
        for i in range(len(x)):
            pts.append(Point.from_dpoint(DPoint(x[i] / dbu, y[i] / dbu)))

        polygon = Polygon(pts)
        shapes(LayerSiN).insert(polygon)

        # Pin on the waveguide:
        from SiEPIC._globals import PIN_LENGTH as pin_length

        x = 0
        t = Trans(Trans.R0, x, 0)
        pin = Path([Point(-pin_length / 2, 0), Point(pin_length / 2, 0)], self.t / dbu)
        pin_t = pin.transformed(t)
        shapes(LayerPinRecN).insert(pin_t)
        text = Text("pin1", t)
        shape = shapes(LayerPinRecN).insert(text)
        shape.text_size = 0.4 / dbu

        # Reference information
        t = Trans(Trans.R0, 0, -4000)
        text = Text(
            "Ref: 'Universal grating coupler design'\nhttps://doi.org/10.1117/12.2042185\nPCell implementation by: Yun Wang, Timothy Richards, Adam DeAbreu,\nJonas Flueckiger, Charlie Lin, Lukas Chrostowski, Connor Mosquera",
            t,
        )
        shape = shapes(LayerPinRecN).insert(text)
        shape.text_size = 0.4 / dbu
        shape.text_halign = 2  # right alignment

        t = Trans(Trans.R0, 0, 4000)
        text = Text(
            "Wavelength: %s\nIncident Angle: %s\nPolarization: %s\nSilicon thickness: %s\nSilicon etch depth: %s"
            % (
                self.wavelength,
                self.theta_c,
                self.pol,
                self.Si_thickness,
                self.etch_depth,
            ),
            t,
        )
        shape = shapes(LayerPinRecN).insert(text)
        shape.text_size = 0.4 / dbu
        shape.text_halign = 2  # right alignment
        shape.text_valign = 2  # bottom alignment

        # Device recognition layer
        yr = sin(start) * (
            N * lambda_0 / (n_e * (1 - e * cos(float(start)))) + gc_number * period
        )
        box1 = Box(
            -(self.grating_length + self.taper_length) / dbu - pin_length * 2,
            yr / dbu,
            0,
            -yr / dbu,
        )
        shapes(LayerDevRecN).insert(box1)



import pya
from pya import *
from SiEPIC.utils import get_technology_by_name
from . import *


class cavity_hole(pya.PCellDeclarationHelper):
    """
    The PCell declaration for the cavity hole.
    This is an example of how to perform inversion operations
    by Jaspreet, Jingda, and Lukas
    """

    def __init__(self):
        # Important: initialize the super class
        super(cavity_hole, self).__init__()
        TECHNOLOGY = get_technology_by_name("EBeam")

        # declare the parameters
        self.param("x", self.TypeDouble, "x coordinate", default=0)
        self.param("y", self.TypeDouble, "y coordinate", default=0)
        self.param("radius", self.TypeDouble, "hole radius", default=10)
        self.param(
            "gap",
            self.TypeDouble,
            "half of the gap between surrounding holes (microns)",
            default=2,
        )
        self.param("LayerSi", self.TypeLayer, "Si Layer", default=TECHNOLOGY["Si"])
        self.param(
            "pinrec", self.TypeLayer, "PinRec Layer", default=TECHNOLOGY["PinRec"]
        )
        self.param(
            "devrec", self.TypeLayer, "DevRec Layer", default=TECHNOLOGY["DevRec"]
        )

    def display_text_impl(self):
        # Provide a descriptive text for the cell
        return "cavity_hole(R=" + ("%.3f-%.3f" % (self.radius, self.gap)) + ")"

    def can_create_from_shape(self, layout, shape, layer):
        return False

    def produce_impl(self):
        dbu = self.layout.dbu
        ly = self.layout
        shapes = self.cell.shapes

        # Layers definitions
        LayerSi = self.layer
        LayerSiN = ly.layer(self.LayerSi)
        LayerPinRecN = ly.layer(self.pinrec)
        LayerDevRecN = ly.layer(self.devrec)
        TECHNOLOGY = get_technology_by_name("EBeam")
        LayerTextN = TECHNOLOGY["Text"]
        # cell: layout cell to place the layout
        # LayerSiN: which layer to use
        # x, y: location of the origin
        # r: radius
        # w: waveguide width
        # length units in dbu

        x = self.x / dbu
        y = self.y / dbu
        r = self.radius / dbu
        r = r / 2
        w = 2 * r

        gap = self.gap / dbu  # gap between one hole to another

        # function to generate points to create a circle
        def circle(x, y, r):
            npts = 32
            theta = 2 * math.pi / npts  # increment, in radians
            pts = []
            for i in range(0, npts):
                pts.append(
                    Point.from_dpoint(
                        DPoint(
                            (x + r * math.cos(i * theta)) / 1,
                            (y + r * math.sin(i * theta)) / 1,
                        )
                    )
                )
            return pts

        # rectangular shapes
        side_gap = gap  # half of gap between surrounding holes
        dx = side_gap + (2 * r)
        dy = side_gap + (2 * r)

        Si_slab = Region()
        Si_slab.insert(Box(-dx / 2, -dy / 2, dx / 2, dy / 2))

        # Circle
        hole = Region()
        hole_cell = circle(0, 0, r)
        hole_poly = Polygon(hole_cell)
        hole_t = hole_poly.transformed(Trans(Trans.R0, x, y))
        hole.insert(hole_t)

        # perform inversion:
        phc = Si_slab - hole
        self.cell.shapes(LayerSiN).insert(phc)
        # inversion can also be done by performing the XOR function
        # upper_poly = xor(Region(Polygon(circleu_pts)),Region(Polygon(rectu_pts)))

        return "cavity_hole(R=" + ("%.3f-%.3f" % (r, gap)) + ")"



# $description: Unit testing: all EBeam libraries' cells
# $show-in-menu
# $group-name: Examples_EBeam
# $menu-path: siepic_menu.exlayout.end
# Unit testing for all library layout fixed cells and PCells

import os
import sys

path_GitHub = os.path.expanduser("~/Documents/GitHub/")
if 0:
    if "SiEPIC" not in sys.modules:
        path_siepic = os.path.join(
            path_GitHub, "SiEPIC-Tools/klayout_dot_config/python"
        )
        if path_siepic not in sys.path:
            sys.path.insert(0, path_siepic)
from SiEPIC._globals import Python_Env

print("KLayout running in mode: %s" % Python_Env)
from SiEPIC.utils.layout import new_layout
from SiEPIC.utils import load_klayout_technology
from SiEPIC.scripts import instantiate_all_library_cells, export_layout

path_module = os.path.join(path_GitHub, "SiEPIC_EBeam_PDK/klayout")
path_lyt_file = os.path.join(path_GitHub, "SiEPIC_EBeam_PDK/klayout/EBeam/EBeam.lyt")
tech = load_klayout_technology("EBeam", path_module, path_lyt_file)


# Create a new layout
topcell, ly = new_layout(tech.name, "UnitTesting", GUI=True, overwrite=True)
# Instantiate all cells
instantiate_all_library_cells(topcell)
# Save the layout
path_out = os.path.dirname(os.path.realpath(__file__))
ly.write(os.path.join(path_out, "a.gds"))
export_layout(topcell, path_out, "a_static", format="oas")
print("done")



import pya
from . import *
from SiEPIC.utils import get_technology_by_name
from SiEPIC._globals import PIN_LENGTH as pin_length
from pya import DPolygon
import math

# ====================================================================================================
# File:     e_skid_single_side_taper.py
# Author:   Hang (Bobby) Zou
# Purpose:  PCell in EBeam library for a common parallel e-skid design that have adjustable period
#           and fill factor
# Version:  V2.0
# Date:     2023-11-09
# ====================================================================================================


class e_skid_single_side_taper(pya.PCellDeclarationHelper):
    def __init__(self):
        # Important: initialize the super class
        super(e_skid_single_side_taper, self).__init__()
        self.technology_name = "EBeam"

        TECHNOLOGY = get_technology_by_name(self.technology_name)

        # e_skid Parameters:
        # core parameter
        self.param(
            "divider0", self.TypeNone, "==============Core Parameter=============="
        )

        self.param("w", self.TypeDouble, "Core Waveguide Width [um]", default=0.3)

        self.param(
            "divider1", self.TypeNone, "==============Clad Parameter=============="
        )

        # clad parameter
        self.param("p", self.TypeDouble, "e-skid Period [um]", default=0.1)
        self.param("ff", self.TypeDouble, "e-skid Fill Factor [0.2-0.8]", default=0.5)
        self.param("p_num", self.TypeInt, "e-skid Period Number", default=5)
        self.param(
            "length", self.TypeDouble, "Length of e-skid Waveguide [um]", default=10
        )

        self.param(
            "divider2", self.TypeNone, "==============Taper Parameter============="
        )

        # taper
        self.param("taper_True", self.TypeBoolean, "Taper [T/F]", default=True)
        self.param(
            "taperL", self.TypeDouble, "Strip -> e-skid Taper Length [um]", default=5
        )
        self.param("taperW", self.TypeDouble, "Strip Width [um]", default=0.5)

        self.param(
            "divider3", self.TypeNone, "============Extended Clad Parameter=========="
        )

        # extended clad parameter
        self.param(
            "clad_taper_L", self.TypeDouble, "Extended e-skid Length [um]", default=5
        )
        self.param(
            "clad_taper_angle",
            self.TypeDouble,
            "Extended e-skid Angle [degree]",
            default=2,
        )

        self.param("divider4", self.TypeNone, "=====================================")

        # Layer Parameters - Don't touch
        self.param("layer", self.TypeLayer, "Layer", default=TECHNOLOGY["Si"])
        self.param(
            "pinrec", self.TypeLayer, "PinRec Layer", default=TECHNOLOGY["PinRec"]
        )
        self.param(
            "devrec", self.TypeLayer, "DevRec Layer", default=TECHNOLOGY["DevRec"]
        )
        self.param(
            "oxideopen",
            self.TypeLayer,
            "Oxide Open Layer",
            default=TECHNOLOGY["Oxide open (to BOX)"],
        )

    def display_text_impl(self):
        # Provide a descriptive text for the cell
        return "e_skid_with_a_single_side_taper"

    # Creating the layout
    def produce_impl(self):
        # Layout Parameters
        dbu = self.layout.dbu
        ly = self.layout
        shapes = self.cell.shapes

        LayerSi = self.layer
        LayerSiN = ly.layer(LayerSi)
        LayerPinRecN = ly.layer(self.pinrec)
        LayerDevRecN = ly.layer(self.devrec)

        w = self.w  # Core Waveguide Width [um]
        p = self.p  # e-skid Period [um]
        ff = self.ff  # e-skid Duty Cycle [%]
        p_num = self.p_num  # e-skid Period Number
        length = self.length  # e-skid Waveguide Length [um]

        taper_True = self.taper_True  # e-skid Taper [T/F]
        taperW = self.taperW  # e-skid Taper Width [um]
        taperL = self.taperL  # e-skid Taper Length [um]

        clad_taper_L = self.clad_taper_L  # Extended e-skid length [nm]
        clad_taper_angle_radians = (
            self.clad_taper_angle * math.pi / 180
        )  # Extended e-skid angle [degrees]

        ###############################
        ########### e-skid ############
        ###############################

        # 1. e-skid main body

        # x, y coordinates of the core
        x_core = [0, length, length, 0]
        y_core = [-w / 2, -w / 2, w / 2, w / 2]

        dpts = [
            pya.DPoint(x_core[i], y_core[i]) for i in range(len(x_core))
        ]  # Core body
        dpolygon = DPolygon(dpts)
        element = Polygon.from_dpoly(dpolygon * (1 / dbu))
        shapes(LayerSiN).insert(element)

        # draw multiple clad for both top and bottom
        for i in range(1, p_num + 1):
            # x, y coordinates of the top clad
            x_clad_top = [0, length, length, 0]
            y_clad_top = [
                w / 2 + i * p - (p * ff),
                w / 2 + i * p - (p * ff),
                w / 2 + i * p,
                w / 2 + i * p,
            ]

            dpts = [
                pya.DPoint(x_clad_top[i], y_clad_top[i]) for i in range(len(x_clad_top))
            ]  # Core body
            dpolygon = DPolygon(dpts)
            element = Polygon.from_dpoly(dpolygon * (1 / dbu))
            shapes(LayerSiN).insert(element)

            # x, y coordinates of the bottom clad
            x_clad_bottom = [0, length, length, 0]
            y_clad_bottom = [
                -w / 2 - i * p + (p * ff),
                -w / 2 - i * p + (p * ff),
                -w / 2 - i * p,
                -w / 2 - i * p,
            ]

            dpts = [
                pya.DPoint(x_clad_bottom[i], y_clad_bottom[i])
                for i in range(len(x_clad_bottom))
            ]  # Core body
            dpolygon = DPolygon(dpts)
            element = Polygon.from_dpoly(dpolygon * (1 / dbu))
            shapes(LayerSiN).insert(element)

        # 2. taper + Pins

        if taper_True:
            # Top clad
            if clad_taper_L > taperL:
                # Prevent the clad_taper_L to be larger than taperL
                clad_taper_L = taperL

            # Calculate the required offset
            w_offset = taperW * clad_taper_L / taperL

            # fetch the taper angle at the set point
            taper_angle = math.atan(((taperW - w) / 2) / (taperL))

            if clad_taper_angle_radians <= taper_angle:
                clad_taper_angle_radians = taper_angle

            for i in range(1, p_num + 1):
                # draw the left clad, angle method
                # locate the end points of the extended clad
                clad_end_point_position = (
                    w / 2
                    + i * p
                    + i * (clad_taper_L * math.tan(clad_taper_angle_radians))
                )

                # x, y coordinates of the top clad
                x_clad_top = [0, -clad_taper_L, -clad_taper_L, 0]
                y_clad_top = [
                    w / 2 + i * p - (p * ff),
                    clad_end_point_position - (p * ff),
                    clad_end_point_position,
                    w / 2 + i * p,
                ]

                dpts = [
                    pya.DPoint(x_clad_top[i], y_clad_top[i])
                    for i in range(len(x_clad_top))
                ]  # Core body
                dpolygon = DPolygon(dpts)
                element = Polygon.from_dpoly(dpolygon * (1 / dbu))
                shapes(LayerSiN).insert(element)

                # x, y coordinates of the bottom clad
                x_clad_bottom = [0, -clad_taper_L, -clad_taper_L, 0]
                y_clad_bottom = [
                    -w / 2 - i * p + (p * ff),
                    -clad_end_point_position + (p * ff),
                    -clad_end_point_position,
                    -w / 2 - i * p,
                ]

                dpts = [
                    pya.DPoint(x_clad_bottom[i], y_clad_bottom[i])
                    for i in range(len(x_clad_bottom))
                ]  # Core body
                dpolygon = DPolygon(dpts)
                element = Polygon.from_dpoly(dpolygon * (1 / dbu))
                shapes(LayerSiN).insert(element)

            # left taper
            x_taper_left = [0, 0, -taperL, -taperL]
            y_taper_left = [-w / 2, w / 2, taperW / 2, -taperW / 2]

            dpts = [
                pya.DPoint(x_taper_left[i], y_taper_left[i])
                for i in range(len(x_taper_left))
            ]  # Core body
            dpolygon = DPolygon(dpts)
            element = Polygon.from_dpoly(dpolygon * (1 / dbu))
            shapes(LayerSiN).insert(element)

            # Adding Pins for mating
            # Pin 1
            t = pya.DTrans(pya.Trans.R90, -taperL / dbu, 0 / dbu)
            pin = pya.Path(
                [pya.DPoint(0, -pin_length / 2), pya.DPoint(0, pin_length / 2)],
                taperW / dbu,
            )
            pin_t = pin.transformed(t)
            shapes(LayerPinRecN).insert(pin_t)
            text = pya.Text("opt1", pya.DTrans(pya.Trans.R0, -taperL / dbu, 0 / dbu))
            shape = shapes(LayerPinRecN).insert(text)
            shape.text_size = 0.2 / dbu

            # Pin 2
            t = pya.DTrans(pya.Trans.R270, (length) / dbu, 0 / dbu)
            pin = pya.Path(
                [pya.DPoint(0, -pin_length / 2), pya.DPoint(0, pin_length / 2)], w / dbu
            )
            pin_t = pin.transformed(t)
            shapes(LayerPinRecN).insert(pin_t)
            text = pya.Text("opt2", pya.DTrans(pya.Trans.R0, (length) / dbu, 0 / dbu))
            shape = shapes(LayerPinRecN).insert(text)
            shape.text_size = 0.2 / dbu

            # Devbox
            dev = pya.DBox(
                -taperL,
                clad_end_point_position + 0.5,
                length,
                -clad_end_point_position - 0.5,
            )

            shapes(LayerDevRecN).insert(dev)

        else:
            # Adding Pins for mating
            # Pin 1
            t = pya.DTrans(pya.Trans.R90, 0, 0 / dbu)
            pin = pya.Path(
                [pya.DPoint(0, -pin_length / 2), pya.DPoint(0, pin_length / 2)], w / dbu
            )
            pin_t = pin.transformed(t)
            shapes(LayerPinRecN).insert(pin_t)
            text = pya.Text("opt1", pya.DTrans(pya.Trans.R0, 0, 0 / dbu))
            shape = shapes(LayerPinRecN).insert(text)
            shape.text_size = 0.2 / dbu

            # Pin 2
            t = pya.DTrans(pya.Trans.R270, length / dbu, 0 / dbu)
            pin = pya.Path(
                [pya.DPoint(0, -pin_length / 2), pya.DPoint(0, pin_length / 2)], w / dbu
            )
            pin_t = pin.transformed(t)
            shapes(LayerPinRecN).insert(pin_t)
            text = pya.Text("opt2", pya.DTrans(pya.Trans.R0, length / dbu, 0 / dbu))
            shape = shapes(LayerPinRecN).insert(text)
            shape.text_size = 0.2 / dbu

            # Devbox
            dev = pya.DBox(0, w / 2 + p_num * p + 0.5, length, -w / 2 - p_num * p - 0.5)

            shapes(LayerDevRecN).insert(dev)


# ====================================================================================================
# END OF DOCUMENT
# ====================================================================================================



import pya
from pya import *
from SiEPIC.utils import get_technology_by_name


class Tapered_Ring(pya.PCellDeclarationHelper):
    """
    The PCell declaration for the tapered ring.
    Author: Mustafa Hammood     mustafa@ece.ubc.ca
    """

    def __init__(self):
        # Important: initialize the super class
        super(Tapered_Ring, self).__init__()
        TECHNOLOGY = get_technology_by_name("EBeam")

        # declare the parameters
        self.param("w_top", self.TypeDouble, "Top width", default=0.5)
        self.param("w_bot", self.TypeDouble, "Bottom width (>Top Width)", default=2)
        self.param("radius", self.TypeDouble, "Radius", default=10)
        self.param("LayerSi", self.TypeLayer, "Si Layer", default=TECHNOLOGY["Si"])
        self.param(
            "pinrec", self.TypeLayer, "PinRec Layer", default=TECHNOLOGY["PinRec"]
        )
        self.param(
            "devrec", self.TypeLayer, "DevRec Layer", default=TECHNOLOGY["DevRec"]
        )

    def display_text_impl(self):
        # Provide a descriptive text for the cell
        return "Tapered_Ring(R=" + ("%.3f-%.3f" % (self.radius, self.w_bot)) + ")"

    def can_create_from_shape(self, layout, shape, layer):
        return False

    def produce_impl(self):
        from pya import Region, Polygon
        import math

        # fetch parameters
        dbu = self.layout.dbu
        ly = self.layout
        shapes = self.cell.shapes

        # Layers definitions
        LayerSi = self.layer
        LayerSiN = ly.layer(self.LayerSi)
        LayerPinRecN = ly.layer(self.pinrec)
        LayerDevRecN = ly.layer(self.devrec)
        # LayerTextN = ly.layer(LayerText)

        # cell: layout cell to place the layout
        # LayerSiN: which layer to use
        # x, y: location of the origin
        # r: radius
        # w: waveguide width
        # length units in dbu

        x = 0
        w_top = self.w_top / dbu
        r = self.radius / dbu

        w_bot = self.w_bot / dbu

        deltaW = w_bot - w_top

        nptsFactor = 12

        # function to generate points to create circle
        def circle(x, w_top, r):
            npts = 32 * nptsFactor
            theta = 2 * math.pi / npts  # increment, in radians
            pts = []
            for i in range(0, npts):
                pts.append(
                    Point.from_dpoint(
                        DPoint(
                            (x + r * math.cos(i * theta)) / 1,
                            (w_top + r * math.sin(i * theta)) / 1,
                        )
                    )
                )
            return pts

        # function to generate points to create innercircle
        def inner_circle(x, w_top, r):
            npts = 32 * nptsFactor
            theta = 2 * math.pi / npts  # increment, in radians
            pts = []
            for i in range(0, npts):
                pts.append(
                    Point.from_dpoint(
                        DPoint(
                            (x + r * math.cos(i * theta)) / 1,
                            (w_top + r * math.sin(i * theta)) / 1,
                        )
                    )
                )
            return pts

        # Outer circle
        x = 0
        y = 0
        r_out = r + w_top / 2
        ring = Region()
        ring_cell = circle(0, 0, r_out)
        ring_poly = Polygon(ring_cell)
        # ring_t = ring_poly.transformed(Trans(x,y))
        ring.insert(ring_poly)

        # Inner erasing circle
        r_in = r - (w_top) / 2 - deltaW / 2
        x = 0
        y = -deltaW / 2
        # Inner Circle
        hole = Region()
        hole_cell = inner_circle(x, y, r_in)
        hole_poly = Polygon(hole_cell)
        # hole_t = hole_poly.transformed(Trans(x,y))
        hole.insert(hole_poly)

        # perform inversion:
        phc = ring - hole
        self.cell.shapes(LayerSiN).insert(phc)
        # inversion can also be done by performing the XOR function
        # upper_poly = xor(Region(Polygon(circleu_pts)),Region(Polygon(rectu_pts)))

        return "Tapered_Ring(R=" + ("%.3f-%.3f" % (r, w_bot)) + ")"



import pya
from pya import *
from SiEPIC.utils import get_technology_by_name


class Waveguide_SWG_Assist(pya.PCellDeclarationHelper):
    """
    Input: length, target_period, wg_width, width, duty
    """

    def __init__(self):
        # Important: initialize the super class
        super(Waveguide_SWG_Assist, self).__init__()
        TECHNOLOGY = get_technology_by_name("EBeam")

        # declare the parameters
        self.param("length", self.TypeDouble, "Waveguide length", default=10.0)
        self.param(
            "target_period", self.TypeDouble, "Target period (microns)", default=0.200
        )
        self.param("swg_wg_width", self.TypeDouble, "SWG Waveguide width", default=0.5)
        self.param(
            "strip_wg_width", self.TypeDouble, "Strip Waveguide width", default=0.06
        )
        self.param("duty", self.TypeDouble, "Duty Cycle (0 to 1)", default=0.5)
        self.param("layer", self.TypeLayer, "Layer", default=TECHNOLOGY["Si"])
        self.param(
            "pinrec", self.TypeLayer, "PinRec Layer", default=TECHNOLOGY["PinRec"]
        )
        self.param(
            "devrec", self.TypeLayer, "DevRec Layer", default=TECHNOLOGY["DevRec"]
        )

    #    self.param("textl", self.TypeLayer, "Text Layer", default = LayerInfo(10, 0))

    def display_text_impl(self):
        # Provide a descriptive text for the cell
        return "Waveguide_SWG_Assist_%s-%.3f-%.3f-%.3f-%.3f" % (
            self.length,
            self.target_period,
            self.swg_wg_width,
            self.strip_wg_width,
            self.duty,
        )

    def coerce_parameters_impl(self):
        pass

    def can_create_from_shape(self, layout, shape, layer):
        return False

    def produce_impl(self):
        # fetch the parameters
        dbu = self.layout.dbu
        ly = self.layout
        shapes = self.cell.shapes

        LayerSi = self.layer
        LayerSiN = ly.layer(LayerSi)
        # LayerSiSPN = ly.layer(LayerSiSP)
        LayerPinRecN = ly.layer(self.pinrec)
        LayerDevRecN = ly.layer(self.devrec)

        # Determine the period such that the waveguide length is as desired.  Slight adjustment to period
        N_boxes = int(round(self.length / self.target_period - 0.5))
        grating_period = self.length / (N_boxes) / dbu
        # print("N boxes: %s, grating_period: %s" % (N_boxes, grating_period) )

        # Draw the Bragg grating:
        box_width = int(round(grating_period * self.duty))

        w = self.swg_wg_width / dbu
        half_w = w / 2
        for i in range(0, N_boxes + 1):
            x = int(round((i * grating_period - box_width / 2)))
            box1 = Box(x, -half_w, x + box_width, half_w)
            shapes(LayerSiN).insert(box1)
        length = self.length / dbu

        # Strip waveguide:
        w_strip = self.strip_wg_width / dbu
        if w_strip > 0:
            box1 = Box(0, -w_strip / 2, length, w_strip / 2)
            shapes(LayerSiN).insert(box1)

        # Pins on the waveguide:
        from SiEPIC._globals import PIN_LENGTH as pin_length

        #    pin_length = box_width
        t = Trans(Trans.R0, 0, 0)
        pin = Path([Point(pin_length / 2, 0), Point(-pin_length / 2, 0)], w)
        pin_t = pin.transformed(t)
        shapes(LayerPinRecN).insert(pin_t)
        text = Text("pin1", t)
        shape = shapes(LayerPinRecN).insert(text)
        shape.text_size = 0.4 / dbu

        t = Trans(Trans.R0, length, 0)
        pin = Path([Point(-pin_length / 2, 0), Point(pin_length / 2, 0)], w)
        pin_t = pin.transformed(t)
        shapes(LayerPinRecN).insert(pin_t)
        text = Text("pin2", t)
        shape = shapes(LayerPinRecN).insert(text)
        shape.text_halign = 2
        shape.text_size = 0.4 / dbu

        # Compact model information
        t = Trans(Trans.R0, 0, 0)
        text = Text("Lumerical_INTERCONNECT_library=Design kits/ebeam", t)
        shape = shapes(LayerDevRecN).insert(text)
        shape.text_size = 0.1 / dbu
        t = Trans(Trans.R0, length / 10, 0)
        text = Text("Component=NO_MODEL_AVAILABLE", t)
        shape = shapes(LayerDevRecN).insert(text)
        shape.text_size = 0.1 / dbu
        t = Trans(Trans.R0, length / 9, -box_width * 2)
        text = Text(
            "Spice_param:length=%.3fu target_period=%.3fu grating_period=%.3fu swg_wg_width=%.3fu strip_wg_width=%.3fu duty=%.3f "
            % (
                self.length,
                self.target_period,
                round(grating_period) * dbu,
                self.swg_wg_width,
                self.strip_wg_width,
                self.duty,
            ),
            t,
        )
        shape = shapes(LayerDevRecN).insert(text)
        shape.text_size = 0.1 / dbu

        # Create the device recognition layer -- make it 1 * wg_width away from the waveguides.
        points = [pya.Point(0, 0), pya.Point(length, 0)]
        path = pya.Path(points, w)
        path = pya.Path(points, w * 3)
        shapes(LayerDevRecN).insert(path.simple_polygon())



import pya
from pya import *
from SiEPIC.utils import get_technology_by_name


class MultiBox_BraggGrating(pya.PCellDeclarationHelper):
    """
    Reference:
    Enxiao Luan, Han Yun, Minglei Ma, Daniel M. Ratner, Karen C. Cheung, and Lukas Chrostowski,
    "Label-free biosensing with a multi-box sub-wavelength phase-shifted Bragg grating waveguide"
    Biomedical Optics Express Vol. 10, Issue 9, pp. 4825-4838 (2019)
    https://doi.org/10.1364/BOE.10.004825
    Multibox Bragg Grating
    By Enxiao Luan at UBC, 2020, with fix by Lukas Chrostowski
    """

    def __init__(self):
        # Important: initialize the super class
        super(MultiBox_BraggGrating, self).__init__()
        TECHNOLOGY = get_technology_by_name("EBeam")

        # declare the parameters
        self.param("pitch", self.TypeDouble, "Period of SWG [um]", default=0.24)
        self.param("w", self.TypeDouble, "Waveguide Width [um]", default=0.18)
        self.param(
            "wc", self.TypeDouble, "Corrugation width [um]", default=0.1
        )  # minimum radius is 5 um in TE
        self.param("ff", self.TypeDouble, "Duty Cycle", default=0.75)
        # self.param("angle", self.TypeDouble, "Angle", default = 360)
        self.param("gap", self.TypeDouble, "Row Gap [um]", default=0.06)
        # self.param("gap2", self.TypeDouble, "Bus Gap [um]", default = 0.06)
        self.param("row", self.TypeDouble, "Number of Rows of main WG", default=3)
        self.param(
            "Length",
            self.TypeDouble,
            "Bragg Grating Length on Each side [um]",
            default=30,
        )
        self.param("taperL", self.TypeDouble, "Taper Length [um]", default=10)
        self.param(
            "phaseshifted",
            self.TypeBoolean,
            "Bragg Type (Uniform=False, Phase-shifted=True)",
            default=False,
        )
        self.param("layer", self.TypeLayer, "Layer", default=TECHNOLOGY["Si"])
        self.param(
            "pinrec", self.TypeLayer, "PinRec Layer", default=TECHNOLOGY["PinRec"]
        )
        self.param(
            "devrec", self.TypeLayer, "DevRec Layer", default=TECHNOLOGY["DevRec"]
        )

    #    self.param("textl", self.TypeLayer, "Text Layer", default = LayerInfo(10, 0))

    def display_text_impl(self):
        # Provide a descriptive text for the cell
        return "MultiBox_BraggGrating_length%.3f um " % (2 * self.Length)

    def coerce_parameters_impl(self):
        pass

    def can_create_from_shape(self, layout, shape, layer):
        return False

    def produce_impl(self):
        from SiEPIC.extend import to_itype
        from pya import DPolygon
        import math
        # This is the main part of the implementation: create the layout

        # fetch the parameters
        dbu = self.layout.dbu
        ly = self.layout
        shapes = self.cell.shapes
        LayerSi = self.layer
        LayerSiN = ly.layer(LayerSi)
        LayerPinRecN = ly.layer(self.pinrec)
        LayerDevRecN = ly.layer(self.devrec)

        # variables
        pitch = self.pitch
        w = self.w
        wc = self.wc
        ff = self.ff
        # angle = self.angle
        gap = self.gap
        # gap2 = self.gap2
        row = self.row
        Length = self.Length
        taperL = self.taperL
        phaseshifted = self.phaseshifted
        # print(doublebus)

        if Length <= 3 * pitch:
            Length = 3 * pitch
            print(
                "invalid length of MultiBox Bragg grating, set length at least 3 times larger than the period of multibox blocks"
            )

        s1 = pitch * ff  # silicon
        s2 = pitch - s1  # gap

        # Draw the first Multi-box Waveguides
        # calulate ideal length of bus
        Bragg_length = Length
        # bus_length = self.cell.bbox().height()*dbu +pitch*2
        constant = math.ceil(Bragg_length / (s1 + s2))
        if Bragg_length % (s1 + s2) != 0:
            Bragg_length = constant * (s1 + s2)

        for ii in range(0, int(row)):
            xo = [
                (ii * (w + gap)),
                (w + ii * (w + gap)),
                (w + ii * (w + gap)),
                (ii * (w + gap)),
            ]
            yo = [0, 0, s1, s1]
            dpts = [pya.DPoint(xo[i], yo[i]) for i in range(len(xo))]
            dpolygon = DPolygon(dpts)
            element = Polygon.from_dpoly(dpolygon * (1.0 / dbu))
            shapes(LayerSiN).insert(element)

            # draw the SWG waveguide from the center one to +/- positions
            for i in range(0, int(math.ceil((constant) / 2))):
                yu = [yo[j] + i * pitch for j in range(len(yo))]
                yd = [yo[j] - i * pitch for j in range(len(yo))]

                dpts = [pya.DPoint(xo[i], yu[i]) for i in range(len(xo))]
                dpolygon = DPolygon(dpts)
                element = Polygon.from_dpoly(dpolygon * (1.0 / dbu))
                shapes(LayerSiN).insert(element)
                dpts = [pya.DPoint(xo[i], yd[i]) for i in range(len(xo))]
                dpolygon = DPolygon(dpts)
                element = Polygon.from_dpoly(dpolygon * (1.0 / dbu))
                shapes(LayerSiN).insert(element)

            # draw the tapers from waveguide to SWG
            if taperL != 0:
                xtu = [
                    ((ii * (w + gap)) + (w - 0.06) / 2),
                    ((ii * (w + gap)) + (w - 0.06) / 2 + 0.06),
                    ((ii * (w + gap)) + w + gap / 2),
                    ((ii * (w + gap)) - gap / 2),
                ]
                ytu = [(yu[3] - taperL), (yu[3] - taperL), (yu[3]), (yu[3])]
                dpts = [pya.DPoint(xtu[i], ytu[i]) for i in range(len(xtu))]
                dpolygon = DPolygon(dpts)
                element = Polygon.from_dpoly(dpolygon * (1.0 / dbu))
                shapes(LayerSiN).insert(element)

                ytd = [(yd[1] + taperL), (yd[1] + taperL), (yd[1]), (yd[1])]
                dpts = [pya.DPoint(xtu[i], ytd[i]) for i in range(len(xtu))]
                dpolygon = DPolygon(dpts)
                element = Polygon.from_dpoly(dpolygon * (1.0 / dbu))
                shapes(LayerSiN).insert(element)

                # taper connections from multi-tapers to std WG
                xTu = [
                    (-gap / 2),
                    (-gap / 2 + row * (gap + w)),
                    ((row * (gap + w) - gap) / 2 + 0.25),
                    ((row * (gap + w) - gap) / 2 - 0.25),
                ]
                yTu = [(yu[3]), (yu[3]), (yu[3] + taperL), (yu[3] + taperL)]
                dpts = [pya.DPoint(xTu[i], yTu[i]) for i in range(len(xTu))]
                dpolygon = DPolygon(dpts)
                element = Polygon.from_dpoly(dpolygon * (1.0 / dbu))
                shapes(LayerSiN).insert(element)

                yTd = [(yd[1]), (yd[1]), (yd[1] - taperL), (yd[1] - taperL)]
                dpts = [pya.DPoint(xTu[i], yTd[i]) for i in range(len(xTu))]
                dpolygon = DPolygon(dpts)
                element = Polygon.from_dpoly(dpolygon * (1.0 / dbu))
                shapes(LayerSiN).insert(element)
                # f is the factor to define the Dev starting point
                f = 1

            if taperL == 0:
                yTd = yd
                yTu = yu
                xTu = [0, 0, 0, 0]
                f = 0

        # draw the corrugations of the Bragg gratings
        constant_c = math.ceil((Bragg_length - taperL) / (s1 + s2))

        for iii in range(0, int(math.ceil(constant_c / 4))):
            yuc = [yo[j] + 2 * iii * pitch for j in range(len(yo))]
            ydc = [yo[j] - 2 * iii * pitch for j in range(len(yo))]
            xlc = [-wc - s2, -s2, -s2, -wc - s2]
            xrc = [row * pitch, row * pitch + wc, row * pitch + wc, row * pitch]

            dpts = [pya.DPoint(xlc[i], yuc[i]) for i in range(len(xo))]
            dpolygon = DPolygon(dpts)
            element = Polygon.from_dpoly(dpolygon * (1.0 / dbu))
            shapes(LayerSiN).insert(element)

            dpts = [pya.DPoint(xrc[i], yuc[i]) for i in range(len(xo))]
            dpolygon = DPolygon(dpts)
            element = Polygon.from_dpoly(dpolygon * (1.0 / dbu))
            shapes(LayerSiN).insert(element)

        if phaseshifted is False:
            for iii in range(0, int(math.ceil(constant_c / 4))):
                # yuc = [yo[j]+2*iii*pitch for j in range(len(yo))]
                ydc = [yo[j] - 2 * iii * pitch for j in range(len(yo))]
                # xlc = [-wc-s2, -s2, -s2,-wc-s2]
                # xrc = [row*pitch, row*pitch+wc, row*pitch+wc, row*pitch]

                dpts = [pya.DPoint(xlc[i], ydc[i]) for i in range(len(xo))]
                dpolygon = DPolygon(dpts)
                element = Polygon.from_dpoly(dpolygon * (1.0 / dbu))
                shapes(LayerSiN).insert(element)

                dpts = [pya.DPoint(xrc[i], ydc[i]) for i in range(len(xo))]
                dpolygon = DPolygon(dpts)
                element = Polygon.from_dpoly(dpolygon * (1.0 / dbu))
                shapes(LayerSiN).insert(element)

        if phaseshifted is True:
            for iii in range(0, int(math.ceil(constant_c / 4))):
                # yuc = [yo[j]+2*iii*pitch for j in range(len(yo))]
                yddc = [yo[j] - pitch - 2 * iii * pitch for j in range(len(yo))]

                dpts = [pya.DPoint(xlc[i], yddc[i]) for i in range(len(xo))]
                dpolygon = DPolygon(dpts)
                element = Polygon.from_dpoly(dpolygon * (1.0 / dbu))
                shapes(LayerSiN).insert(element)

                dpts = [pya.DPoint(xrc[i], yddc[i]) for i in range(len(xo))]
                dpolygon = DPolygon(dpts)
                element = Polygon.from_dpoly(dpolygon * (1.0 / dbu))
                shapes(LayerSiN).insert(element)

        # DEV BOX
        dev = Box(
            f * (-gap - wc - (2) * pitch) / dbu,
            yTd[-1] / dbu,
            (wc + (row + 2) * pitch) / dbu,
            yTu[-1] / dbu,
        )
        shapes(LayerDevRecN).insert(dev)
        dev_width = self.cell.bbox().width() / 2

        dev_up = yu[-1] / dbu + taperL / dbu
        dev_down = yd[-4] / dbu - taperL / dbu

        # Create the pins on the waveguides, as short paths:
        from SiEPIC._globals import PIN_LENGTH as pin_length

        w = to_itype(self.w, dbu)
        gap = to_itype(self.gap, dbu)

        # Pin1
        t = Trans(Trans.R0, xTu[-1] / dbu + 250, dev_up)
        pin = Path(
            [Point(0, -pin_length / 2), Point(0, pin_length / 2)], 500
        )  # 500 is width of std WG
        pin_t = pin.transformed(t)
        shapes(LayerPinRecN).insert(pin_t)
        t = Trans(Trans.R270, xTu[-1] / dbu + 250, dev_up)
        text = Text("pin1", t)
        text.halign = 0
        shape = shapes(LayerPinRecN).insert(text)
        shape.text_size = 0.4 / dbu

        # Pin2
        t = Trans(Trans.R0, xTu[-1] / dbu + 250, dev_down)
        pin = Path([Point(0, pin_length / 2), Point(0, -pin_length / 2)], 500)
        pin_t = pin.transformed(t)
        shapes(LayerPinRecN).insert(pin_t)
        t = Trans(Trans.R90, xTu[-1] / dbu + 250, dev_down)
        text = Text("pin2", t)
        text.halign = 0
        shape = shapes(LayerPinRecN).insert(text)
        shape.text_size = 0.4 / dbu

        # Ref
        t = Trans(Trans.R270, xTu[-1] / dbu + 250 + 0.5 / dbu, dev_up)
        text = Text("Ref: E. Luan, doi.org/10.1364/BOE.10.004825", t)
        text.halign = 0
        shape = shapes(LayerPinRecN).insert(text)
        shape.text_size = 0.4 / dbu



import pya
from pya import *
import os
import pathlib

# import pcells_EBeam ### folder name ###
import importlib

# Equivalent to from custom_exceptions_path import *
current_directory = os.path.dirname(os.path.abspath(__file__))
custom_exceptions_path = os.path.join(current_directory, "custom_exceptions.py")
exec(open(custom_exceptions_path).read())

"""
Python script to test that all EBeam libraries are registered and all Pcells are properly registered in their respective library and 
will display polygons when placed on a new layout.

To run this script from the command line: '{path to klayout} -zz -r {path to this script/EBeam_Lib_PCellTests.py}'.

This script can also be run within KLayout.

This is a preliminary test script and will be implemented using GitHub Actions to be run every time a new Pcell is added to the code base.

by Jasmina Brar
2023/08

== Change log ==
2023/10/30, Jasmina Brar.
    - raise exceptions when errors occur and exits under these conditions
    - tests successful pcell registration in all libraries 

"""

library_folders = [
    "pcells_EBeam",
    "pcells_EBeam_Beta",
    "pcells_SiN",
    "pcells_EBeam_Dream",
]
library_names = ["EBeam", "EBeam_Beta", "EBeam-SiN", "EBeam-Dream"]
tech_names = ["EBeam", "EBeam", "EBeam", "EBeam"]

for i in range(len(library_folders)):
    # get all .py files in library folder
    files = [
        f
        for f in os.listdir(
            os.path.join(
                os.path.dirname(os.path.realpath(__file__)), library_folders[i]
            )
        )
        if ".py" in pathlib.Path(f).suffixes and "__init__" not in f
    ]

    importlib.invalidate_caches()

    # get library layout object
    tech_name = tech_names[i]
    library_name = library_names[i]
    library = pya.Library().library_by_name(library_name, tech_name)

    print("*** Testing library: %s" % library_name)

    # If library does not exist in klayout
    if library == None:
        raise LibraryNotRegistered(library_name)

    layout = library.layout()

    # check that the library is registered in klayout
    try:
        if layout == None:
            raise LibraryNotRegistered(library_name)

    except LibraryNotRegistered as e:
        print("Caught {}: {}".format(type(e).__name__, str(e)))
        pya.Application.instance().exit(1)

    # loop through all pcells from this library's folder
    for f in files:
        try:
            mm = f.replace(".py", "")

            print("  * Testing cell: %s" % mm)

            # check that the pcell has been registered in the library's layout
            if mm not in layout.pcell_names():
                raise PCellRegistrationError(mm, library_name)

            # instantiate pcell in a new layer and check that it contains polygons
            new_layout = pya.Layout()
            pcell_decl = layout.pcell_declaration(mm)
            new_layout.register_pcell(mm, pcell_decl)

            parameter_decl = pcell_decl.get_parameters()

            all_params = {}
            for p in parameter_decl:
                all_params[p.name] = p.default

            pcell = new_layout.create_cell(mm, all_params)

            # check that there were no errors generated from the pcell

            error_shapes = pcell.shapes(new_layout.error_layer())

            for error in error_shapes.each():
                raise PCellImplementationError(mm, library_name, error.text)

            if pcell.is_empty() or pcell.bbox().area() == 0:
                raise PCellInstantiationError(mm, library_name)

            # topcell = new_layout.create_cell("top")
            # t = Trans(Trans.R0, 0,0)
            # inst = topcell.insert(CellInstArray(pcell.cell_index(), t))

        except (PCellRegistrationError, PCellInstantiationError, Exception) as e:
            print("Caught {}: {}".format(type(e).__name__, str(e)))
            pya.Application.instance().exit(1)

    print(
        "Complete. All pcells from {} folder were successfully registered in {} library".format(
            library_folders[i], library_names[i]
        )
    )

print("Complete. All pcells were succcessfully registered in all libraries.")



"""
This file is part of the SiEPIC_EBeam_PDK
by Lukas Chrostowski, et al., (c) 2023

This file implements a library called "EBeam-SiN", for silicon nitride


Version history:

Lukas Chrostowski, 2023/01/24
 - Creation of a new library for Silicon Nitride devices
 - directional coupler PCell for 895 nm
 - Added SiN waveguide to WAVEGUIDES.XML

Lukas Chrostowski, 2023/11
 - components for 1550 and 1310
 - waveguides for 1550 and 1310
 - compound waveguide
 - compatibility with PyPI usage of KLayout

"""

verbose = False

if verbose:
    print("siepic_ebeam_library_SiN")

from pya import *
import os
import pathlib
import sys

dir_path = os.path.dirname(os.path.realpath(__file__))
if dir_path not in sys.path:
    sys.path.append(dir_path)

files = [
    f
    for f in os.listdir(
        os.path.join(os.path.dirname(os.path.realpath(__file__)), "pcells_SiN")
    )
    if ".py" in pathlib.Path(f).suffixes and "__init__" not in f
]
import pcells_SiN  ### folder name ###
import importlib

importlib.invalidate_caches()
pcells_ = []
for f in files:
    module = "pcells_SiN.%s" % f.replace(".py", "")  ### folder name ###
    if verbose:
        print(" - found module: %s" % module)
    m = importlib.import_module(module)
    if verbose:
        print(m)
    pcells_.append(importlib.reload(m))


class siepic_ebeam_library_hubbard(Library):
    """
    The library where we will put the PCells and GDS into
    """

    def __init__(self):
        tech_name = "EBeam"
        library = tech_name + "-SiN"
        self.technology = tech_name

        # Set the description
        self.description = "v0.4.22, Silicon Nitride"

        if verbose:
            print("Initializing '%s' Library, %s" % (library, self.description))

        # Save the path, used for loading WAVEGUIDES.XML
        import os

        self.path = os.path.dirname(os.path.realpath(__file__))

        # Import all the GDS files from the tech folder
        import os
        import fnmatch

        dir_path = os.path.normpath(
            os.path.join(
                os.path.dirname(os.path.realpath(__file__)), "../gds/EBeam_SiN"
            )
        )
        if verbose:
            print("  library path: %s" % dir_path)
        search_str = "*.[Oo][Aa][Ss]"  # OAS
        for root, dirnames, filenames in os.walk(dir_path, followlinks=True):
            for filename in fnmatch.filter(filenames, search_str):
                file1 = os.path.join(root, filename)
                if verbose:
                    print(" - reading %s" % file1)
                self.layout().read(file1)
        search_str = "*.[Gg][Dd][Ss]"  # GDS
        for root, dirnames, filenames in os.walk(dir_path, followlinks=True):
            for filename in fnmatch.filter(filenames, search_str):
                file1 = os.path.join(root, filename)
                if verbose:
                    print(" - reading %s" % file1)
                self.layout().read(file1)

        # Create the PCell declarations
        for m in pcells_:
            mm = m.__name__.replace("pcells_SiN.", "")
            mm2 = m.__name__ + "." + mm + "()"
            if verbose:
                print(" - register_pcell %s, %s" % (mm, mm2))
            self.layout().register_pcell(mm, eval(mm2))
        if verbose:
            print(" done with pcells")

        # Register us the library with the technology name
        # If a library with that name already existed, it will be replaced then.
        self.register(library)


# Instantiate and register the library
siepic_ebeam_library_hubbard()



from pya import *
import pya


class Waveguide_Straight(pya.PCellDeclarationHelper):
    """
    Input: length, width
    draws a straight waveguide with pins. centred at the instantiation point.
    Usage: instantiate, and use transformations (rotation)
    """

    def __init__(self):
        # Important: initialize the super class
        super(Waveguide_Straight, self).__init__()
        from SiEPIC.utils import get_technology_by_name

        TECHNOLOGY = get_technology_by_name("EBeam")

        # declare the parameters
        self.param("wg_length", self.TypeInt, "Waveguide Length", default=10000)
        self.param("wg_width", self.TypeInt, "Waveguide width", default=500)
        self.param("layer", self.TypeLayer, "Layer", default=TECHNOLOGY["Si"])
        self.param(
            "pinrec", self.TypeLayer, "PinRec Layer", default=TECHNOLOGY["PinRec"]
        )
        self.param(
            "devrec", self.TypeLayer, "DevRec Layer", default=TECHNOLOGY["DevRec"]
        )

    def display_text_impl(self):
        # Provide a descriptive text for the cell
        return "Waveguide_Straight_%.3f-%.3f" % (
            self.wg_length / 1000,
            self.wg_width / 1000,
        )

    def coerce_parameters_impl(self):
        pass

    def can_create_from_shape(self, layout, shape, layer):
        return False

    def produce_impl(self):
        ly = self.layout
        LayerSiN = ly.layer(self.layer)
        LayerPinRecN = ly.layer(self.pinrec)
        LayerDevRecN = ly.layer(self.devrec)

        # fetch the parameters
        dbu = self.layout.dbu
        ly = self.layout
        shapes = self.cell.shapes

        LayerSi = self.layer
        LayerSiN = ly.layer(LayerSi)
        # LayerPinRecN = ly.layer(self.pinrec)
        LayerDevRecN = ly.layer(self.devrec)

        #    print("Waveguide_Straight:")
        w = self.wg_width
        length = self.wg_length
        points = [[-length / 2, 0], [length / 2, 0]]
        path = Path([Point(-length / 2, 0), Point(length / 2, 0)], w)
        #    print(path)

        shapes(LayerSiN).insert(path.simple_polygon())

        from SiEPIC._globals import PIN_LENGTH

        # Pins on the bus waveguide side:
        pin_length = PIN_LENGTH
        if length < pin_length + 1:
            pin_length = int(length / 3)
            pin_length = math.ceil(pin_length / 2.0) * 2
        if pin_length == 0:
            pin_length = 2

        t = Trans(Trans.R0, -length / 2, 0)
        pin = Path([Point(pin_length / 2, 0), Point(-pin_length / 2, 0)], w)
        pin_t = pin.transformed(t)
        shapes(LayerPinRecN).insert(pin_t)
        text = Text("pin1", t)
        shape = shapes(LayerPinRecN).insert(text)
        shape.text_size = 0.4 / dbu

        t = Trans(Trans.R0, length / 2, 0)
        pin = Path([Point(-pin_length / 2, 0), Point(pin_length / 2, 0)], w)
        pin_t = pin.transformed(t)
        shapes(LayerPinRecN).insert(pin_t)
        text = Text("pin2", t)
        shape = shapes(LayerPinRecN).insert(text)
        shape.text_size = 0.4 / dbu
        shape.text_halign = 2

        # Compact model information
        t = Trans(Trans.R0, 0, 0)
        text = Text("Lumerical_INTERCONNECT_library=Design kits/EBeam", t)
        shape = shapes(LayerDevRecN).insert(text)
        shape.text_size = 0.1 / dbu
        t = Trans(Trans.R0, length / 10, 0)
        text = Text("Lumerical_INTERCONNECT_component=ebeam_wg_integral_1550", t)
        shape = shapes(LayerDevRecN).insert(text)
        shape.text_size = 0.1 / dbu
        t = Trans(Trans.R0, length / 9, 0)
        text = Text(
            "Spice_param:wg_width=%.3fu wg_length=%.3fu"
            % (self.wg_width * dbu, self.wg_length * dbu),
            t,
        )
        shape = shapes(LayerDevRecN).insert(text)
        shape.text_size = 0.1 / dbu

        # Create the device recognition layer -- make it 1 * wg_width away from the waveguides.
        path = Path([Point(-length / 2, 0), Point(length / 2, 0)], w * 3)
        shapes(LayerDevRecN).insert(path.simple_polygon())



import pya
from SiEPIC.utils import get_technology_by_name
from pya import *


class spiral(pya.PCellDeclarationHelper):
    """
    Input:
    """

    def __init__(self):
        # Important: initialize the super class
        super(spiral, self).__init__()
        TECHNOLOGY = get_technology_by_name("EBeam")

        # declare the parameters
        self.param("length", self.TypeDouble, "Target Waveguide length", default=10.0)
        self.param(
            "wg_width", self.TypeDouble, "Waveguide width (microns)", default=0.5
        )
        self.param("min_radius", self.TypeDouble, "Minimum radius (microns)", default=5)
        self.param(
            "wg_spacing", self.TypeDouble, "Waveguide spacing (microns)", default=1
        )
        self.param(
            "spiral_ports", self.TypeInt, "Ports on the same side? 0/1", default=0
        )
        self.param("layer", self.TypeLayer, "Layer", default=TECHNOLOGY["Si"])
        self.param(
            "pinrec", self.TypeLayer, "PinRec Layer", default=TECHNOLOGY["PinRec"]
        )
        self.param(
            "devrec", self.TypeLayer, "DevRec Layer", default=TECHNOLOGY["DevRec"]
        )

    def display_text_impl(self):
        # Provide a descriptive text for the cell
        return "spiral_%s-%.3f-%.3f-%.3f" % (
            self.length,
            self.wg_width,
            self.min_radius,
            self.wg_spacing,
        )

    def coerce_parameters_impl(self):
        pass

    def can_create_from_shape(self, layout, shape, layer):
        return False

    def produce_impl(self):
        debug = False
        # fetch the parameters
        dbu = self.layout.dbu
        ly = self.layout
        shapes = self.cell.shapes

        LayerSi = self.layer
        LayerSiN = ly.layer(LayerSi)
        LayerPinRecN = ly.layer(self.pinrec)
        LayerDevRecN = ly.layer(self.devrec)

        # draw spiral
        from math import pi, cos, sin

        # Archimedes spiral
        # r = b + a * theta
        b = self.min_radius
        spacing = self.wg_spacing + self.wg_width
        a = 2 * spacing / (2 * pi)

        # area, length, turn tracking for spiral
        area = 0
        spiral_length = 0
        turn = -1

        from SiEPIC.utils import points_per_circle, arc_wg_xy

        while spiral_length < self.length:
            turn += 1

            # Spiral #1
            pts = []
            # local radius:
            r = 2 * b + a * turn * 2 * pi - self.wg_width / 2
            # number of points per circle:
            npoints = int(points_per_circle(r))
            # increment, in radians, for each point:
            da = 2 * pi / npoints
            # draw the inside edge of spiral
            for i in range(0, npoints + 1):
                t = i * da
                xa = (a * t + r) * cos(t)
                ya = (a * t + r) * sin(t)
                pts.append(Point.from_dpoint(DPoint(xa / dbu, ya / dbu)))
            # draw the outside edge of spiral
            r = 2 * b + a * turn * 2 * pi + self.wg_width / 2
            npoints = int(points_per_circle(r))
            da = 2 * pi / npoints
            for i in range(npoints, -1, -1):
                t = i * da
                xa = (a * t + r) * cos(t)
                ya = (a * t + r) * sin(t)
                pts.append(Point.from_dpoint(DPoint(xa / dbu, ya / dbu)))
            polygon = Polygon(pts)
            area += polygon.area()
            shapes(LayerSiN).insert(polygon)

            # Spiral #2
            pts = []
            # local radius:
            r = 2 * b + a * turn * 2 * pi - self.wg_width / 2 - spacing
            # number of points per circle:
            npoints = int(points_per_circle(r))
            # increment, in radians, for each point:
            da = 2 * pi / npoints
            # draw the inside edge of spiral
            for i in range(0, npoints + 1):
                t = i * da + pi
                xa = (a * t + r) * cos(t)
                ya = (a * t + r) * sin(t)
                pts.append(Point.from_dpoint(DPoint(xa / dbu, ya / dbu)))
            # draw the outside edge of spiral
            r = 2 * b + a * turn * 2 * pi + self.wg_width / 2 - spacing
            npoints = int(points_per_circle(r))
            da = 2 * pi / npoints
            for i in range(npoints, -1, -1):
                t = i * da + pi
                xa = (a * t + r) * cos(t)
                ya = (a * t + r) * sin(t)
                pts.append(Point.from_dpoint(DPoint(xa / dbu, ya / dbu)))
            polygon = Polygon(pts)
            area += polygon.area()
            shapes(LayerSiN).insert(polygon)

            # waveguide length:
            spiral_length = area / self.wg_width * dbu * dbu + 2 * pi * self.min_radius

        if self.spiral_ports:
            # Spiral #1 extra 1/2 arm
            turn = turn + 1
            pts = []
            # local radius:
            r = 2 * b + a * turn * 2 * pi - self.wg_width / 2
            # number of points per circle:
            npoints = int(points_per_circle(r))
            # increment, in radians, for each point:
            da = pi / npoints
            # draw the inside edge of spiral
            for i in range(0, npoints + 1):
                t = i * da
                xa = (a * t + r) * cos(t)
                ya = (a * t + r) * sin(t)
                pts.append(Point.from_dpoint(DPoint(xa / dbu, ya / dbu)))
            # draw the outside edge of spiral
            r = 2 * b + a * turn * 2 * pi + self.wg_width / 2
            npoints = int(points_per_circle(r))
            da = pi / npoints
            for i in range(npoints, -1, -1):
                t = i * da
                xa = (a * t + r) * cos(t)
                ya = (a * t + r) * sin(t)
                pts.append(Point.from_dpoint(DPoint(xa / dbu, ya / dbu)))
            polygon = Polygon(pts)
            area += polygon.area()
            shapes(LayerSiN).insert(polygon)
            turn = turn - 1
            # waveguide length:
            spiral_length = area / self.wg_width * dbu * dbu + 2 * pi * self.min_radius

        # Centre S-shape connecting waveguide
        # layout_arc_wg_dbu(self.cell, LayerSiN, -b/dbu, 0, b/dbu, self.wg_width/dbu, 0, 180)
        self.cell.shapes(LayerSiN).insert(
            arc_wg_xy(-b / dbu, 0, b / dbu, self.wg_width / dbu, 0, 180)
        )
        # layout_arc_wg_dbu(self.cell, LayerSiN, b/dbu, 0, b/dbu, self.wg_width/dbu, 180, 0)
        self.cell.shapes(LayerSiN).insert(
            arc_wg_xy(b / dbu, 0, b / dbu, self.wg_width / dbu, 180, 0)
        )
        if debug:
            print("spiral length: %s microns" % spiral_length)

        # Pins on the waveguide:
        from SiEPIC._globals import PIN_LENGTH as pin_length

        x = -(2 * b + a * (turn + 1) * 2 * pi) / dbu
        w = self.wg_width / dbu
        t = Trans(Trans.R0, x, 0)
        pin = Path([Point(0, pin_length / 2), Point(0, -pin_length / 2)], w)
        pin_t = pin.transformed(t)
        shapes(LayerPinRecN).insert(pin_t)
        text = Text("pin2", t)
        shape = shapes(LayerPinRecN).insert(text)
        shape.text_size = 0.4 / dbu

        if self.spiral_ports:
            x = -(2 * b + a * (turn + 1.5) * 2 * pi) / dbu
        else:
            x = (2 * b + a * (turn + 1) * 2 * pi) / dbu
        t = Trans(Trans.R0, x, 0)
        pin = Path([Point(0, -pin_length / 2), Point(0, pin_length / 2)], w)
        pin_t = pin.transformed(t)
        shapes(LayerPinRecN).insert(pin_t)
        text = Text("pin1", t)
        shape = shapes(LayerPinRecN).insert(text)
        shape.text_size = 0.4 / dbu

        # Compact model information
        t = Trans(Trans.R0, -abs(x), 0)
        text = Text("Length=%.3fu" % spiral_length, t)
        shape = shapes(LayerDevRecN).insert(text)
        shape.text_size = abs(x) / 8
        t = Trans(Trans.R0, 0, 0)
        text = Text("Lumerical_INTERCONNECT_library=Design kits/ebeam_v1.2", t)
        shape = shapes(LayerDevRecN).insert(text)
        shape.text_size = 0.1 / dbu
        t = Trans(Trans.R0, 0, w * 2)
        text = Text("Component=ebeam_wg_strip_1550", t)
        shape = shapes(LayerDevRecN).insert(text)
        shape.text_size = 0.1 / dbu
        t = Trans(Trans.R0, 0, -w * 2)
        text = Text(
            "Spice_param:wg_length=%.3fu wg_width=%.3fu min_radius=%.3fu wg_spacing=%.3fu"
            % (spiral_length, self.wg_width, (self.min_radius), self.wg_spacing),
            t,
        )
        shape = shapes(LayerDevRecN).insert(text)
        shape.text_size = 0.1 / dbu

        # Create the device recognition layer -- make it 1 * wg_width away from the waveguides.
        x = abs(x)
        npoints = int(points_per_circle(x) / 10)
        da = 2 * pi / npoints  # increment, in radians
        r = x + 2 * self.wg_width / dbu
        pts = []
        for i in range(0, npoints + 1):
            pts.append(Point.from_dpoint(DPoint(r * cos(i * da), r * sin(i * da))))
        shapes(LayerDevRecN).insert(Polygon(pts))

        # print("spiral done.")



import pya
from pya import *
from SiEPIC.utils import get_technology_by_name


class ebeam_rib_phase_shifter_folded(pya.PCellDeclarationHelper):
    # N-doped heater phase shifter in a rib waveguide
    # design by Guowu Zhang <guowu.zhang@mail.mcgill.ca>
    # and Lukas Chrostowski
    # based on Ref: R. Priti, "Efficiency Improvement of an O-band SOI-MZI Thermo-optic Matrix Switch"
    # and folding based on
    # Ref: Z. Lu, "Michelson Interferometer Thermo-optic Switch on SOI with a 50 uW Power Consumption", IEEE PTL, 2015

    def __init__(self):
        # Important: initialize the super class
        super(ebeam_rib_phase_shifter_folded, self).__init__()
        TECHNOLOGY = get_technology_by_name("EBeam")

        # declare the parameters
        self.param("silayer", self.TypeLayer, "Si Layer", default=TECHNOLOGY["Si"])
        self.param(
            "io_wg_type",
            self.TypeBoolean,
            "I/O waveguide type, 0 - strip; 1 - rib",
            default=0,
        )
        self.param(
            "fold_number", self.TypeInt, "Number of folds (3, 5, 7, ...)", default=5
        )
        self.param(
            "input_rib_width", self.TypeDouble, "Input Rib Width (um)", default=0.35
        )
        self.param(
            "input_slab_width", self.TypeDouble, "Input Slab Width (um)", default=2
        )
        self.param(
            "in_taper_length", self.TypeDouble, "Input Taper Length (um)", default=15
        )
        self.param("length", self.TypeDouble, "Phase Shifter Length (um)", default=150)
        self.param("width", self.TypeDouble, "Waveguide Width (um)", default=0.5)
        #    self.param("ps_slab_width", self.TypeDouble, "Phase Shifter Slab Width (um)", default = 18)
        self.param("npp_width", self.TypeDouble, "Npp Width (um)", default=2)
        self.param(
            "npp_distance", self.TypeDouble, "Npp Edge to Si Edge (um)", default=2
        )
        self.param("segments", self.TypeInt, "Phase Shifter Segments", default=5)
        self.param(
            "overlay",
            self.TypeDouble,
            "Overlay accuracy (optical litho) (um)",
            default=1,
        )
        self.param(
            "overlay_ebl", self.TypeDouble, "Overlay accuracy (EBL) (um)", default=0.05
        )
        #    self.param("vc_dw", self.TypeDouble, "VC Offset", default = 5)
        self.param("vc_w", self.TypeDouble, "Contact via width", default=5)
        #    self.param("m_dw", self.TypeDouble, "Metal Offset", default = 4)
        self.param("m_w", self.TypeDouble, "Metal Width", default=10)
        self.param(
            "slayer", self.TypeLayer, "Slab Layer", default=TECHNOLOGY["Si - 90 nm rib"]
        )
        #    self.param("nlayer", self.TypeLayer, "N Layer", default = TECHNOLOGY['Si N'])
        self.param(
            "npplayer", self.TypeLayer, "N++ Layer", default=TECHNOLOGY["Si N++"]
        )
        self.param("vclayer", self.TypeLayer, "VC Layer", default=TECHNOLOGY["VC"])
        self.param(
            "mlayer", self.TypeLayer, "Metal Layer", default=TECHNOLOGY["M2_router"]
        )
        self.param(
            "pinrec", self.TypeLayer, "PinRec Layer", default=TECHNOLOGY["PinRec"]
        )
        self.param(
            "devrec", self.TypeLayer, "DevRec Layer", default=TECHNOLOGY["DevRec"]
        )
        self.param("textl", self.TypeLayer, "Text Layer", default=LayerInfo(10, 0))

    def display_text_impl(self):
        # Provide a descriptive text for the cell
        return "ebeam_rib_phase_shifter_folded(L=" + ("%.3f" % self.length) + ")"

    def produce(self, layout, layers, parameters, cell):
        # coerce parameters (make consistent)

        self._layers = layers
        self.cell = cell
        self._param_values = parameters
        self.layout = layout

        # cell: layout cell to place the layout
        # LayerSiN: which layer to use
        # r: radius
        # w: waveguide width
        # length units in dbu

        from SiEPIC._globals import PIN_LENGTH
        from SiEPIC.utils.layout import layout_waveguide2
        from SiEPIC.extend import to_itype

        # fetch the parameters
        TECHNOLOGY = get_technology_by_name("EBeam")
        dbu = self.layout.dbu
        ly = self.layout

        LayerSiN = ly.layer(self.silayer)
        LayerSlab = ly.layer(self.slayer)
        #    LayerNN = ly.layer(self.nlayer)
        LayerNPPN = ly.layer(self.npplayer)
        LayerVCN = ly.layer(self.vclayer)
        LayerMN = ly.layer(self.mlayer)
        LayerPinRecN = ly.layer(self.pinrec)
        LayerDevRecN = ly.layer(self.devrec)
        TextLayerN = ly.layer(self.textl)

        ## define parameters for phase shifter
        overlay = to_itype(self.overlay, dbu)
        overlay_ebl = to_itype(self.overlay_ebl, dbu)

        ## define waveguide related parameters
        radius_um, adiab, bezier = 5, 1, 0.2
        w = to_itype(self.width, dbu)  # waveguide width, default 0.5 um
        l = to_itype(self.length, dbu)  # pahse shifter length, default 150 um
        npp_d = to_itype(self.npp_distance, dbu)  # npp to center waveguide distance
        sw = (npp_d + w / 2) * 2  # twice of npp_d that is defined above
        in_rib = to_itype(
            self.input_rib_width, dbu
        )  # the slab width of the input/output waveguide to the phase shifter
        in_taper = to_itype(
            self.in_taper_length, dbu
        )  # the input waveguide to phase shifter dimension transfer taper length, default 15
        edge_slab_width = to_itype(
            4, dbu
        )  # width of the 90 nm slab that is outside of all the folding waveguide, default 2 * 2um
        npp_width = to_itype(self.npp_width, dbu)  # npp width, default 5 um

        ## define waveguide folding parameters
        folding_n = int(
            round((self.fold_number - 1) / 2)
        )  # folding waveguide number on one side
        folding_w1 = w + 50  # one of the folding waveguide width
        folding_w2 = w - 50  # the other one of the folding waveguide width
        folding_gap = to_itype(2, dbu)  # the gap between two folding waveguide
        routing_step = to_itype(
            radius_um, dbu
        )  # define how large the U shape should be, default 20

        ## define node related parameters
        contact_size = to_itype(self.vc_w, dbu)  # square NPP size, for VC purpose
        pin_text_size = to_itype(0.4, dbu)  # pin text size

        metal_routing_distance_to_node = to_itype(
            30, dbu
        )  # *** calculate from PCell parameters ***
        metal_routing_width = to_itype(self.m_w, dbu)  # width of metal
        vc_to_npp_exclusion = to_itype(
            self.overlay, dbu
        )  # VC boundary to NPP boundary distance

        ps_sl_w = (
            sw + 2 * contact_size + folding_n * 2 * (w + folding_gap + overlay_ebl)
        )  # overall pahse shifter 90nm slab width

        # measure the total length of the waveguides
        total_waveguide_length = 0

        ## define strip to slab transfer taper parameters
        N = 100  # Number of points for the input/output slab taper
        order = 3  # input/output slab taper curve

        pts = [Point(-l / 2, 0), Point(l / 2, 0)]
        total_waveguide_length += layout_waveguide2(
            TECHNOLOGY,
            self.layout,
            self.cell,
            ["Si"],
            [w * dbu],
            [0],
            pts,
            radius_um,
            adiab,
            bezier,
        )

        wg2 = pya.Box(-l / 2, -ps_sl_w / 2, l / 2, ps_sl_w / 2)
        wg3 = pya.Box(-l / 2, npp_d + w / 2, l / 2, npp_d + w / 2 + npp_width)
        wg4 = pya.Box(-l / 2, -npp_d - w / 2, l / 2, -npp_d - w / 2 - npp_width)

        self.cell.shapes(LayerNPPN).insert(wg3)
        self.cell.shapes(LayerNPPN).insert(wg4)

        if self.io_wg_type:
            in_slab = to_itype(self.input_slab_width, dbu)
        else:
            in_slab = in_rib

        if folding_n > 0:
            ps_sl_w = (
                contact_size * 2
                + sw * 2
                + (folding_w1 + folding_w2 + folding_gap * 2) * folding_n
                + edge_slab_width
            )
            wg2 = pya.Box(-l / 2, -ps_sl_w / 2, l / 2, ps_sl_w / 2)
            self.cell.shapes(LayerSlab).insert(wg2)

            pts = [
                Point(-l / 2 - in_taper, -w / 2),
                Point(-l / 2, -w / 2),
                Point(-l / 2, w / 2),
                Point(-l / 2 - in_taper, w / 2),
            ]  # fix waveguide width mismatch for left center input taper
            self.cell.shapes(LayerSiN).insert(Polygon(pts))

            # add input slab taper
            pts = []
            for i in range(0, N + 1):
                pts.append(
                    Point(
                        -l / 2 - in_taper + in_taper / N * i,
                        w / 2 + ((sw / 2 + contact_size - w) / (N**order)) * (i**order),
                    )
                )
            for i in range(0, N + 1):
                pts.append(
                    Point(
                        -l / 2 - in_taper + in_taper / N * (N - i),
                        -w / 2
                        - ((sw / 2 + contact_size - w) / (N**order))
                        * ((N - i) ** order),
                    )
                )
            self.cell.shapes(LayerSlab).insert(Polygon(pts))

            # add output strip taper
            pts = [
                Point(l / 2 + in_taper, -w / 2),
                Point(l / 2, -w / 2),
                Point(l / 2, w / 2),
                Point(l / 2 + in_taper, w / 2),
            ]
            # wg2 = pya.Box(-l/2,-4/2/dbu,l/2,4/2/dbu)
            self.cell.shapes(LayerSiN).insert(Polygon(pts))
            total_waveguide_length += in_taper * dbu

            # add input slab taper
            pts = []
            for i in range(0, N + 1):
                pts.append(
                    Point(
                        l / 2 + in_taper - in_taper / N * i,
                        (w - overlay_ebl) / 2
                        + ((sw / 2 + contact_size - (w - overlay_ebl)) / (N**order))
                        * (i**order),
                    )
                )
            for i in range(0, N + 1):
                pts.append(
                    Point(
                        l / 2 + in_taper / N * i,
                        -(w - overlay_ebl) / 2
                        - ((sw / 2 + contact_size - (w - overlay_ebl)) / (N**order))
                        * ((N - i) ** order),
                    )
                )
            self.cell.shapes(LayerSlab).insert(Polygon(pts))

            for i in range(0, folding_n):
                y_coordinate = (
                    contact_size + sw + 2 * vc_to_npp_exclusion
                )  # move up/down the position according to vc_to_npp_exclusion
                y_move = (
                    i * (folding_w1 + folding_w2 + folding_gap * 2) / 2 + y_coordinate
                )
                if i % 2:
                    folding_w = folding_w2
                else:
                    folding_w = folding_w1

                # waveguide in the heated region
                pts = [Point(-l / 2, y_move), Point(l / 2, y_move)]
                total_waveguide_length += layout_waveguide2(
                    TECHNOLOGY,
                    self.layout,
                    self.cell,
                    ["Si"],
                    [folding_w * dbu],
                    [0],
                    pts,
                    radius_um,
                    adiab,
                    bezier,
                )
                pts = [Point(-l / 2, -y_move), Point(l / 2, -y_move)]
                total_waveguide_length += layout_waveguide2(
                    TECHNOLOGY,
                    self.layout,
                    self.cell,
                    ["Si"],
                    [folding_w * dbu],
                    [0],
                    pts,
                    radius_um,
                    adiab,
                    bezier,
                )

                # add upper left input taper
                pts = [
                    Point(-l / 2 - in_taper, -w / 2 + y_move),
                    Point(-l / 2, -folding_w / 2 + y_move),
                    Point(-l / 2, folding_w / 2 + y_move),
                    Point(-l / 2 - in_taper, w / 2 + y_move),
                ]
                # wg2 = pya.Box(-l/2,-4/2/dbu,l/2,4/2/dbu)
                self.cell.shapes(LayerSiN).insert(Polygon(pts))
                total_waveguide_length += in_taper * dbu

                # slab cubic taper
                pts = []
                for ii in range(0, N + 1):
                    pts.append(
                        Point(
                            -l / 2 - in_taper + in_taper / N * ii,
                            (w - overlay_ebl) / 2
                            + ((2 / dbu - (w - overlay_ebl)) / (N**order)) * (ii**order)
                            + y_move,
                        )
                    )
                for ii in range(0, N + 1):
                    pts.append(
                        Point(
                            -l / 2 - in_taper + in_taper / N * (N - ii),
                            -(w - overlay_ebl) / 2
                            - ((2 / dbu - (w - overlay_ebl)) / (N**order))
                            * ((N - ii) ** order)
                            + y_move,
                        )
                    )
                self.cell.shapes(LayerSlab).insert(Polygon(pts))

                if i != folding_n - 1:
                    # add below left taper
                    y_move = -y_move
                    pts = [
                        Point(-l / 2 - in_taper, -w / 2 + y_move),
                        Point(-l / 2, -folding_w / 2 + y_move),
                        Point(-l / 2, folding_w / 2 + y_move),
                        Point(-l / 2 - in_taper, w / 2 + y_move),
                    ]
                    # wg2 = pya.Box(-l/2,-4/2/dbu,l/2,4/2/dbu)
                    self.cell.shapes(LayerSiN).insert(Polygon(pts))
                    total_waveguide_length += in_taper * dbu

                    # slab cubic taper
                    pts = []
                    for ii in range(0, N + 1):
                        pts.append(
                            Point(
                                -l / 2 - in_taper + in_taper / N * ii,
                                (w - overlay_ebl) / 2
                                + ((2 / dbu - (w - overlay_ebl)) / (N**order))
                                * (ii**order)
                                + y_move,
                            )
                        )
                    for ii in range(0, N + 1):
                        pts.append(
                            Point(
                                -l / 2 - in_taper + in_taper / N * (N - ii),
                                -(w - overlay_ebl) / 2
                                - ((2 / dbu - (w - overlay_ebl)) / (N**order))
                                * ((N - ii) ** order)
                                + y_move,
                            )
                        )
                    self.cell.shapes(LayerSlab).insert(Polygon(pts))

                    # add upper right taper
                    y_move = -y_move
                    pts = [
                        Point(l / 2 + in_taper, -w / 2 + y_move),
                        Point(l / 2, -folding_w / 2 + y_move),
                        Point(l / 2, folding_w / 2 + y_move),
                        Point(l / 2 + in_taper, w / 2 + y_move),
                    ]
                    # wg2 = pya.Box(-l/2,-4/2/dbu,l/2,4/2/dbu)
                    self.cell.shapes(LayerSiN).insert(Polygon(pts))
                    total_waveguide_length += in_taper * dbu

                    # slab cubic taper
                    pts = []
                    for ii in range(0, N + 1):
                        pts.append(
                            Point(
                                l / 2 + in_taper - in_taper / N * ii,
                                (w - overlay_ebl) / 2
                                + ((2 / dbu - (w - overlay_ebl)) / (N**order))
                                * (ii**order)
                                + y_move,
                            )
                        )
                    for ii in range(0, N + 1):
                        pts.append(
                            Point(
                                l / 2 + in_taper / N * ii,
                                -(w - overlay_ebl) / 2
                                - ((2 / dbu - (w - overlay_ebl)) / (N**order))
                                * ((N - ii) ** order)
                                + y_move,
                            )
                        )
                    self.cell.shapes(LayerSlab).insert(Polygon(pts))

                else:
                    y_move = -y_move
                    # add input strip taper
                    pts = [
                        Point(-l / 2 - in_taper, -w / 2 + y_move),
                        Point(-l / 2, -folding_w / 2 + y_move),
                        Point(-l / 2, folding_w / 2 + y_move),
                        Point(-l / 2 - in_taper, w / 2 + y_move),
                    ]
                    # wg2 = pya.Box(-l/2,-4/2/dbu,l/2,4/2/dbu)
                    self.cell.shapes(LayerSiN).insert(Polygon(pts))
                    total_waveguide_length += in_taper * dbu

                    # add input slab taper
                    pts = []
                    for i in range(0, N + 1):
                        pts.append(
                            Point(
                                -l / 2 - in_taper + in_taper / N * i,
                                in_slab / 2
                                + ((edge_slab_width - in_slab) / (N**order))
                                * (i**order)
                                + y_move,
                            )
                        )
                    for i in range(0, N + 1):
                        pts.append(
                            Point(
                                -l / 2 - in_taper + in_taper / N * (N - i),
                                -in_slab / 2
                                - ((edge_slab_width - in_slab) / (N**order))
                                * ((N - i) ** order)
                                + y_move,
                            )
                        )
                    self.cell.shapes(LayerSlab).insert(Polygon(pts))

                    # add output strip taper
                    y_move = -y_move
                    pts = [
                        Point(l / 2 + in_taper, -w / 2 + y_move),
                        Point(l / 2, -folding_w / 2 + y_move),
                        Point(l / 2, folding_w / 2 + y_move),
                        Point(l / 2 + in_taper, w / 2 + y_move),
                    ]
                    # wg2 = pya.Box(-l/2,-4/2/dbu,l/2,4/2/dbu)
                    self.cell.shapes(LayerSiN).insert(Polygon(pts))
                    total_waveguide_length += in_taper * dbu

                    # add input slab taper
                    pts = []
                    for i in range(0, N + 1):
                        pts.append(
                            Point(
                                l / 2 + in_taper - in_taper / N * i,
                                in_slab / 2
                                + ((edge_slab_width - in_slab) / (N**order))
                                * (i**order)
                                + y_move,
                            )
                        )
                    for i in range(0, N + 1):
                        pts.append(
                            Point(
                                l / 2 + in_taper / N * i,
                                -in_slab / 2
                                - ((edge_slab_width - in_slab) / (N**order))
                                * ((N - i) ** order)
                                + y_move,
                            )
                        )
                    self.cell.shapes(LayerSlab).insert(Polygon(pts))

                    # Pin on the left side:
                    wg_start = -(l / 2 + in_taper + routing_step * folding_n * 3 + w)
                    p1 = [
                        Point(wg_start + PIN_LENGTH / 2, -y_move),
                        Point(wg_start - PIN_LENGTH / 2, -y_move),
                    ]
                    p1c = Point(wg_start, -y_move)
                    self.set_p1 = p1c
                    self.p1 = p1c
                    pin = Path(p1, w)
                    self.cell.shapes(LayerPinRecN).insert(pin)
                    t = Trans(Trans.R0, wg_start, -y_move)
                    text = Text("pin1", t)
                    shape = self.cell.shapes(LayerPinRecN).insert(text)
                    shape.text_size = pin_text_size
                    # Waveguide to route to port, new in SiEPIC-Tools v0.3.64
                    pts = [Point(-l / 2 - in_taper, -y_move), Point(wg_start, -y_move)]
                    if self.io_wg_type:
                        waveguide_length = layout_waveguide2(
                            TECHNOLOGY,
                            self.layout,
                            self.cell,
                            ["Si", "Si - 90 nm rib"],
                            [w * dbu, in_slab * dbu],
                            [0, 0],
                            pts,
                            radius_um,
                            adiab,
                            bezier,
                        )
                    else:
                        waveguide_length = layout_waveguide2(
                            TECHNOLOGY,
                            self.layout,
                            self.cell,
                            ["Si"],
                            [w * dbu],
                            [0],
                            pts,
                            radius_um,
                            adiab,
                            bezier,
                        )

                    # Pin & Waveguide on the right side:
                    wg_end = l / 2 + in_taper + routing_step * folding_n * 3 + w
                    p2 = [
                        Point(wg_end - PIN_LENGTH / 2, y_move),
                        Point(wg_end + PIN_LENGTH / 2, y_move),
                    ]
                    p2c = Point(wg_end, y_move)
                    self.set_p2 = p2c
                    self.p2 = p2c
                    pin = Path(p2, w)
                    self.cell.shapes(LayerPinRecN).insert(pin)
                    t = Trans(Trans.R0, wg_end, y_move)
                    text = Text("pin2", t)
                    shape = self.cell.shapes(LayerPinRecN).insert(text)
                    shape.text_size = pin_text_size
                    # Waveguide to route to port, new in SiEPIC-Tools v0.3.64
                    pts = [Point(l / 2 + in_taper, y_move), Point(wg_end, y_move)]
                    if self.io_wg_type:
                        total_waveguide_length += layout_waveguide2(
                            TECHNOLOGY,
                            self.layout,
                            self.cell,
                            ["Si", "Si - 90 nm rib"],
                            [w * dbu, in_slab * dbu],
                            [0, 0],
                            pts,
                            radius_um,
                            adiab,
                            bezier,
                        )
                    else:
                        total_waveguide_length += layout_waveguide2(
                            TECHNOLOGY,
                            self.layout,
                            self.cell,
                            ["Si"],
                            [w * dbu],
                            [0],
                            pts,
                            radius_um,
                            adiab,
                            bezier,
                        )

                # add below right taper
                y_move = -y_move
                pts = [
                    Point(l / 2 + in_taper, -w / 2 + y_move),
                    Point(l / 2, -folding_w / 2 + y_move),
                    Point(l / 2, folding_w / 2 + y_move),
                    Point(l / 2 + in_taper, w / 2 + y_move),
                ]
                # wg2 = pya.Box(-l/2,-4/2/dbu,l/2,4/2/dbu)
                self.cell.shapes(LayerSiN).insert(Polygon(pts))
                total_waveguide_length += in_taper * dbu
                pts = []
                for ii in range(0, N + 1):
                    pts.append(
                        Point(
                            l / 2 + in_taper - in_taper / N * ii,
                            (w - overlay_ebl) / 2
                            + ((2 / dbu - (w - overlay_ebl)) / (N**order)) * (ii**order)
                            + y_move,
                        )
                    )
                for ii in range(0, N + 1):
                    pts.append(
                        Point(
                            l / 2 + in_taper / N * ii,
                            -(w - overlay_ebl) / 2
                            - ((2 / dbu - (w - overlay_ebl)) / (N**order))
                            * ((N - ii) ** order)
                            + y_move,
                        )
                    )
                self.cell.shapes(LayerSlab).insert(Polygon(pts))

            # add metal contact and folded waveguides
            for i in range(-folding_n, folding_n):
                y_coordinate = (
                    contact_size + sw + 2 * vc_to_npp_exclusion
                )  # move up/down the position according to vc_to_npp_exclusion
                if i < 0:
                    y_origin = (i + 1) * (
                        folding_w1 + folding_w2 + folding_gap * 2
                    ) / 2 - y_coordinate
                if i == 0:
                    y_origin = 0
                if i > 0:
                    y_origin = (i - 1) * (
                        folding_w1 + folding_w2 + folding_gap * 2
                    ) / 2 + y_coordinate

                if (i + 1) < 0:
                    y_end = (i + 2) * (
                        folding_w1 + folding_w2 + folding_gap * 2
                    ) / 2 - y_coordinate
                if (i + 1) == 0:
                    y_end = 0
                if (i + 1) > 0:
                    y_end = (
                        i * (folding_w1 + folding_w2 + folding_gap * 2) / 2
                        + y_coordinate
                    )

                # Calculate the Folded waveguide vertices
                if (i + folding_n) % 2:
                    # Waveguides on the left side
                    if y_origin + routing_step * 2 - y_end < 2 * radius_um:
                        y_routing = max(routing_step * 2, y_end - y_origin)
                    else:
                        y_routing = y_end + routing_step * 2
                    path = Path(
                        [
                            Point(-l / 2 - in_taper, y_origin),
                            Point(
                                -l / 2
                                - in_taper
                                - routing_step * (1.5 * folding_n - 1.5 * i + 1.5),
                                y_origin,
                            ),
                            Point(
                                -l / 2
                                - in_taper
                                - routing_step * (1.5 * folding_n - 1.5 * i + 1.5),
                                y_routing,
                            ),
                            Point(
                                -l / 2
                                - in_taper
                                - routing_step * (1.5 * folding_n - 1.5 * i - 0.5),
                                y_routing,
                            ),
                            Point(
                                -l / 2
                                - in_taper
                                - routing_step * (1.5 * folding_n - 1.5 * i - 0.5),
                                y_end,
                            ),
                            Point(-l / 2 - in_taper, y_end),
                        ],
                        0,
                    )

                else:
                    # Waveguides on the right side
                    if y_origin - y_end + routing_step * 2 < 2 * radius_um:
                        y_routing = -max(routing_step * 2, y_end - y_origin)
                    else:
                        y_routing = y_origin - routing_step * 2
                    path = Path(
                        [
                            Point(l / 2 + in_taper, y_origin),
                            Point(
                                l / 2
                                + in_taper
                                + routing_step * (1.5 * folding_n + 1.5 * i + 1),
                                y_origin,
                            ),
                            Point(
                                l / 2
                                + in_taper
                                + routing_step * (1.5 * folding_n + 1.5 * i + 1),
                                y_routing,
                            ),
                            Point(
                                l / 2
                                + in_taper
                                + routing_step * (1.5 * folding_n + 1.5 * i + 3),
                                y_routing,
                            ),
                            Point(
                                l / 2
                                + in_taper
                                + routing_step * (1.5 * folding_n + 1.5 * i + 3),
                                y_end,
                            ),
                            Point(l / 2 + in_taper, y_end),
                        ],
                        0,
                    )

                # Draw the waveguide geometry, new in SiEPIC-Tools v0.3.64
                path.unique_points()
                pts = path.get_points()
                total_waveguide_length += layout_waveguide2(
                    TECHNOLOGY,
                    self.layout,
                    self.cell,
                    ["Si"],
                    [w * dbu],
                    [0],
                    pts,
                    radius_um,
                    adiab,
                    bezier,
                )

        for i in range(0, self.segments + 1):
            wg_temp_upper = pya.Box(
                (l - contact_size) / self.segments * i - l / 2 - vc_to_npp_exclusion,
                sw / 2,
                (l - contact_size) / self.segments * i
                - l / 2
                + contact_size
                + vc_to_npp_exclusion,
                sw / 2 + contact_size + 2 * vc_to_npp_exclusion,
            )
            # self.cell.shapes(LayerNN).insert(wg_temp_upper)
            self.cell.shapes(LayerNPPN).insert(wg_temp_upper)
            #        self.cell.shapes(LayerNN).insert(wg_temp_upper)
            wg_temp_upper = pya.Box(
                (l - contact_size) / self.segments * i
                - l / 2
                - 2 * vc_to_npp_exclusion,
                sw / 2,
                (l - contact_size) / self.segments * i
                - l / 2
                + contact_size
                + 2 * vc_to_npp_exclusion,
                sw / 2 + contact_size + 3 * vc_to_npp_exclusion,
            )
            self.cell.shapes(LayerSlab).insert(wg_temp_upper)

            vc_temp_upper = pya.Box(
                (l - contact_size) / self.segments * i - l / 2,
                sw / 2 + vc_to_npp_exclusion,
                (l - contact_size) / self.segments * i - l / 2 + contact_size,
                sw / 2 + contact_size + vc_to_npp_exclusion,
            )
            self.cell.shapes(LayerVCN).insert(vc_temp_upper)

            wg_temp_lower = pya.Box(
                (l - contact_size) / self.segments * i - l / 2 - vc_to_npp_exclusion,
                -sw / 2,
                (l - contact_size) / self.segments * i
                - l / 2
                + contact_size
                + vc_to_npp_exclusion,
                -sw / 2 - contact_size - 2 * vc_to_npp_exclusion,
            )
            # self.cell.shapes(LayerNN).insert(wg_temp_lower)
            self.cell.shapes(LayerNPPN).insert(wg_temp_lower)
            wg_temp_lower = pya.Box(
                (l - contact_size) / self.segments * i
                - l / 2
                - 2 * vc_to_npp_exclusion,
                -sw / 2,
                (l - contact_size) / self.segments * i
                - l / 2
                + contact_size
                + 2 * vc_to_npp_exclusion,
                -sw / 2 - contact_size - 3 * vc_to_npp_exclusion,
            )
            self.cell.shapes(LayerSlab).insert(wg_temp_lower)

            #        self.cell.shapes(LayerNN).insert(wg_temp_lower)
            vc_temp_lower = pya.Box(
                (l - contact_size) / self.segments * i - l / 2,
                -sw / 2 - vc_to_npp_exclusion,
                (l - contact_size) / self.segments * i - l / 2 + contact_size,
                -sw / 2 - contact_size - vc_to_npp_exclusion,
            )
            self.cell.shapes(LayerVCN).insert(vc_temp_lower)

            metal_temp = pya.Box(
                (l - contact_size) / self.segments * i - l / 2 - vc_to_npp_exclusion,
                -sw / 2
                - contact_size
                - metal_routing_distance_to_node * (i % 2)
                - 2 * vc_to_npp_exclusion,
                (l - contact_size) / self.segments * i
                - l / 2
                + contact_size
                + vc_to_npp_exclusion,
                sw / 2
                + contact_size
                + metal_routing_distance_to_node * ((i + 1) % 2)
                + 2 * vc_to_npp_exclusion,
            )
            self.cell.shapes(LayerMN).insert(metal_temp)

        metal_routing = Path(
            [
                Point(-l / 2, -sw / 2 - metal_routing_distance_to_node - contact_size),
                Point(l / 2, -sw / 2 - metal_routing_distance_to_node - contact_size),
            ],
            metal_routing_width,
        )
        self.cell.shapes(LayerMN).insert(metal_routing)
        metal_routing = Path(
            [
                Point(-l / 2, sw / 2 + metal_routing_distance_to_node + contact_size),
                Point(l / 2, sw / 2 + metal_routing_distance_to_node + contact_size),
            ],
            metal_routing_width,
        )
        self.cell.shapes(LayerMN).insert(metal_routing)

        if folding_n > 0:
            dev = Box(
                wg_start,
                -sw / 2
                - metal_routing_distance_to_node
                - contact_size
                - metal_routing_width,
                wg_end,
                sw / 2
                + metal_routing_distance_to_node
                + contact_size
                + metal_routing_width,
            )
            self.cell.shapes(LayerDevRecN).insert(dev)

        t = Trans(0, False, Point(0, 0))
        text = Text("Waveguide Length=%.3fu" % (total_waveguide_length), t, 3 * w, -1)
        shape = self.cell.shapes(LayerDevRecN).insert(text)



import pya
from SiEPIC.utils import get_technology_by_name
from pya import *

class spiral_SiN(pya.PCellDeclarationHelper):
  """
  Input: 
  """

  def __init__(self):

    # Important: initialize the super class
    super(spiral_SiN, self).__init__()
    TECHNOLOGY = get_technology_by_name('EBeam')

    # declare the parameters
    self.param("length", self.TypeDouble, "Target Waveguide length", default = 500)     
    self.param("wg_width", self.TypeDouble, "Waveguide width (microns)", default = 0.75)     
    self.param("min_radius", self.TypeDouble, "Minimum radius (microns)", default = 60)     
    self.param("wg_spacing", self.TypeDouble, "Waveguide spacing (microns)", default = 1)     
    self.param("spiral_ports", self.TypeInt, "Ports on the same side? 0/1", default = 0)
    self.param("layer", self.TypeLayer, "Layer", default = TECHNOLOGY['SiN'])
    self.param("pinrec", self.TypeLayer, "PinRec Layer", default = TECHNOLOGY['PinRec'])
    self.param("devrec", self.TypeLayer, "DevRec Layer", default = TECHNOLOGY['DevRec'])

  def display_text_impl(self):
    # Provide a descriptive text for the cell
    return "spiral_SiN_%s-%.3f-%.3f-%.3f" % \
    (self.length, self.wg_width, self.min_radius, self.wg_spacing)
  
  def coerce_parameters_impl(self):
    pass

  def can_create_from_shape(self, layout, shape, layer):
    return False
    
  def produce_impl(self):
    debug = False
    # fetch the parameters
    dbu = self.layout.dbu
    ly = self.layout
    shapes = self.cell.shapes

    LayerSi = self.layer
    LayerSiN = ly.layer(LayerSi)
    LayerPinRecN = ly.layer(self.pinrec)
    LayerDevRecN = ly.layer(self.devrec)

    # draw spiral
    from math import pi, cos, sin, log, sqrt
    
    # Archimedes spiral
    # r = b + a * theta
    b = self.min_radius
    spacing = self.wg_spacing+self.wg_width
    a = 2*spacing/(2*pi)

    # area, length, turn tracking for spiral
    area = 0
    spiral_length = 0
    turn = -1

    from SiEPIC.utils import points_per_circle, arc_wg_xy

    while spiral_length < self.length:
      turn +=1

      # Spiral #1
      pts = []
      # local radius:
      r = 2*b + a * turn * 2 * pi - self.wg_width/2
      # number of points per circle:
      npoints = int(points_per_circle(r))
      # increment, in radians, for each point:
      da = 2 * pi / npoints  
      # draw the inside edge of spiral
      for i in range(0, npoints+1):
        t = i*da
        xa = (a*t + r) * cos(t)
        ya = (a*t + r) * sin(t)
        pts.append(Point.from_dpoint(DPoint(xa/dbu, ya/dbu)))
      # draw the outside edge of spiral
      r = 2*b + a * turn * 2 * pi + self.wg_width/2
      npoints = int(points_per_circle(r))
      da = 2 * pi / npoints  
      for i in range(npoints, -1, -1):
        t = i*da
        xa = (a*t + r) * cos(t)
        ya = (a*t + r) * sin(t)
        pts.append(Point.from_dpoint(DPoint(xa/dbu, ya/dbu)))
      polygon = Polygon(pts)
      area += polygon.area()
      shapes(LayerSiN).insert(polygon)

      # Spiral #2
      pts = []
      # local radius:
      r = 2*b + a * turn * 2 * pi - self.wg_width/2 - spacing
      # number of points per circle:
      npoints = int(points_per_circle(r))
      # increment, in radians, for each point:
      da = 2 * pi / npoints  
      # draw the inside edge of spiral
      for i in range(0, npoints+1):
        t = i*da + pi
        xa = (a*t + r) * cos(t)
        ya = (a*t + r) * sin(t)
        pts.append(Point.from_dpoint(DPoint(xa/dbu, ya/dbu)))
      # draw the outside edge of spiral
      r = 2*b + a * turn * 2 * pi + self.wg_width/2  - spacing
      npoints = int(points_per_circle(r))
      da = 2 * pi / npoints  
      for i in range(npoints, -1, -1):
        t = i*da + pi
        xa = (a*t + r) * cos(t)
        ya = (a*t + r) * sin(t)
        pts.append(Point.from_dpoint(DPoint(xa/dbu, ya/dbu)))
      polygon = Polygon(pts)
      area += polygon.area()
      shapes(LayerSiN).insert(polygon)

      # waveguide length:
      spiral_length = area / self.wg_width * dbu*dbu + 2 * pi * self.min_radius

    if self.spiral_ports:
      # Spiral #1 extra 1/2 arm
      turn = turn + 1
      pts = []
      # local radius:
      r = 2*b + a * turn * 2 * pi - self.wg_width/2
      # number of points per circle:
      npoints = int(points_per_circle(r))
      # increment, in radians, for each point:
      da = pi / npoints  
      # draw the inside edge of spiral
      for i in range(0, npoints+1):
        t = i*da
        xa = (a*t + r) * cos(t)
        ya = (a*t + r) * sin(t)
        pts.append(Point.from_dpoint(DPoint(xa/dbu, ya/dbu)))
      # draw the outside edge of spiral
      r = 2*b + a * turn * 2 * pi + self.wg_width/2
      npoints = int(points_per_circle(r))
      da = pi / npoints  
      for i in range(npoints, -1, -1):
        t = i*da
        xa = (a*t + r) * cos(t)
        ya = (a*t + r) * sin(t)
        pts.append(Point.from_dpoint(DPoint(xa/dbu, ya/dbu)))
      polygon = Polygon(pts)
      area += polygon.area()
      shapes(LayerSiN).insert(polygon)
      turn = turn - 1
      # waveguide length:
      spiral_length = area / self.wg_width * dbu*dbu + 2 * pi * self.min_radius

    # Centre S-shape connecting waveguide        
    #layout_arc_wg_dbu(self.cell, LayerSiN, -b/dbu, 0, b/dbu, self.wg_width/dbu, 0, 180)
    self.cell.shapes(LayerSiN).insert(arc_wg_xy(-b/dbu, 0, b/dbu, self.wg_width/dbu, 0, 180))
    #layout_arc_wg_dbu(self.cell, LayerSiN, b/dbu, 0, b/dbu, self.wg_width/dbu, 180, 0)
    self.cell.shapes(LayerSiN).insert(arc_wg_xy(b/dbu, 0, b/dbu, self.wg_width/dbu, 180, 0))
    if debug:
        print("spiral length: %s microns" % spiral_length) 

    # Pins on the waveguide:
    from SiEPIC._globals import PIN_LENGTH as pin_length
    
    x = -(2*b + a * (turn+1) * 2 * pi)/dbu
    w = self.wg_width / dbu
    t = Trans(Trans.R0, x,0)
    pin = Path([Point(0,pin_length/2), Point(0,-pin_length/2)], w)
    pin_t = pin.transformed(t)
    shapes(LayerPinRecN).insert(pin_t)
    text = Text ("pin2", t)
    shape = shapes(LayerPinRecN).insert(text)
    shape.text_size = 0.4/dbu

    if self.spiral_ports:
      x = -(2*b + a * (turn+1.5) * 2 * pi)/dbu
    else:
      x = (2*b + a * (turn+1) * 2 * pi)/dbu
    t = Trans(Trans.R0, x,0)
    pin = Path([Point(0,-pin_length/2), Point(0,pin_length/2)], w)
    pin_t = pin.transformed(t)
    shapes(LayerPinRecN).insert(pin_t)
    text = Text ("pin1", t)
    shape = shapes(LayerPinRecN).insert(text)
    shape.text_size = 0.4/dbu

    # Compact model information
    t = Trans(Trans.R0, -abs(x), 0)
    text = Text ('Length=%.3fu' % spiral_length, t)
    shape = shapes(LayerDevRecN).insert(text)
    shape.text_size = abs(x)/8
    t = Trans(Trans.R0, 0, 0)
    text = Text ('Lumerical_INTERCONNECT_library=Design kits/ebeam_v1.2', t)
    shape = shapes(LayerDevRecN).insert(text)
    shape.text_size = 0.1/dbu
    t = Trans(Trans.R0, 0, w*2)
    text = Text ('Component=ebeam_wg_strip_1550', t)
    shape = shapes(LayerDevRecN).insert(text)
    shape.text_size = 0.1/dbu
    t = Trans(Trans.R0, 0, -w*2)
    text = Text \
      ('Spice_param:wg_length=%.3fu wg_width=%.3fu min_radius=%.3fu wg_spacing=%.3fu' %\
      (spiral_length, self.wg_width, (self.min_radius), self.wg_spacing), t )
    shape = shapes(LayerDevRecN).insert(text)
    shape.text_size = 0.1/dbu

    # Create the device recognition layer -- make it 1 * wg_width away from the waveguides.
    x = abs(x)
    npoints = int(points_per_circle(x) /10 )
    da = 2 * pi / npoints # increment, in radians
    r=x + 2 * self.wg_width/dbu
    pts = []
    for i in range(0, npoints+1):
      pts.append(Point.from_dpoint(DPoint(r*cos(i*da), r*sin(i*da))))
    shapes(LayerDevRecN).insert(Polygon(pts))

    # print("spiral done.")
